################################################################################
# Never Eat Yellow Snow APPS, AW939
#
# Helper command line tool for the map_display app. Used to register maps with point lists.
# This application needs scipy and PIL/pillow installed. It was tested with python 3.2.x but should also
# run on python 2.7.x.
#
# Build with pyinstaller, self compiled
#
# Changelog:
#
# V1.0: Initial version
#
# V1.1: - if possible, use the map png's supplied with ac itself to display the map
#           this needs an initial registering process with the script register_map.py
#       - cars are shown as arrows now (if they move) pointing in the direction of their velocity vector
#       - colors of opponent cars changed according to car state (ahead, behind, overlapped, ...)
#
# V1.4: - use .ini file format instead of .reg.pickled
#
# V1.6: - add possibility (code only) for extended fits
#       - runtime optimizations
#       - add possibility to generate .png files from .pickled files
#
# V1.9: - add option for specifying the .png size
#
# V2.9: - refactor line drawing
#
# V2.10:- add sector and drs indicators to generated map
#
# V3.0: - implement meter based scaling and adaptive resolution calculation
#
# V3.3: - implement batch image generation mod
#
# V3.4: - improve the memory usage of create_png.exe
#         - will draw maps in separate tiles, one after another to have less in memory in parallel
#
# V3.7: - support comments in DRS zones and sector definitions
#
# V4.0: - make mark scaling configurable
#       - support subdirs in output path, separated by slash
#       - ignore invalid nodes in fast_lane.ai and try to continue, might fix some tracks
#       - rewrite batch implementation to have usual parameters instead of .ini input
#       - removed pickled files functionality as it was unused
#
# V4.3: - implement circle fit mode to use max space within a rotatable circle
#
# V4.4: - detect close tracks with up to 100m distance (was 30m) between last and first point
#
# V4.5: - add -f flag to force generate the map image, even if hashes might be up to date
#
################################################################################

import sys
import glob
import os.path
import traceback
import numpy
import numpy.linalg
import math
import random
import struct
import configparser
import json
import hashlib
from multiprocessing import cpu_count, freeze_support, Pool
from PIL import Image
from PIL import ImageDraw

create_png_version="4.5" # version of map_display of when create_png was updated. Triggers image regeneration if updated
usage = """Usage:
create_png.py [-s <png_size_px>] [-a <px_per_meter>] [-px | -m] [-c] [-t <included_tracks>] [-fai <fastlane_ai_file> <out_file> | [-bf | -bm] <ac_tracks_path> <output_dir>] <track_width> <outline_width> <outline_color_hex> <fill_color_hex> <pit_color_hex> <mark_scale> <mark_outline_width> <start_fin_color> <sector_color> <drs_color> <drs_detection_color>:
    Config Parameters:
        -s <png_size_px>: is the width and height of the created png file (default 512)
        -a <px_per_meter>: generates an image with an adaptive size based on the track.
                The value defines how many pixels on the image contain one meter on the track.
                Higher means better resolution. (default off, fixed size)
        -px: pixel based scaling mode. Widths are pixels on the final image
        -m : meter based scaling mode. Widths are meters on the track. Without -s the image will be scaled to 1px/m.
        -c : scale map to fit in a circle within the quadratic image to allow rotation without overlap
        -cc: scale map to fit in a circle while still being centered in mass
        -t <included_tracks>: (Batch only) a list of comma separated track folder names from the tracks directory to be used for map generation.
                If set, other tracks from the folder are ignored.

        -fai: (has to be last)
            <fastlane_ai_file> in case of "-fai" the fast_lane.ai file containing the track
                (if it contains a '*', then a globbing scheme will be applied)
            <out_file> out file name (without file extension)
        -bf: (has to be last) (Batch) generate full maps only
        -bm: (has to be last) (Batch) generate mini maps only
            <ac_tracks_path> relative or absolute path to the assettocorsa/content/tracks directory to parse the map data from
            <output_dir> relative or absolute path to the map_display/data directory
        -f: force re-generation even if no parameters have changed in comparison with the existing file
    Required Parameters:
        <track_width> is the track width in pixel
        <outline_width> is the outline width in pixel
        <outline_color_hex> is the hex color code (see below) of the track
        <fill_color_hex> is the hex color code (see below) of the inside of the track
        <pit_color_hex> is the hex color code (see below) of the inside of the pit lane
            (If alpha is 0 it will not be drawn at all)
        <mark_scale> scaling factor for all marks on the map (1.0 would be default scaling)
        <mark_outline_width> outline width for marks on the map in pixel
        <start_fin_color> is the hex color code (see below) of the start/finish indicator
            (If alpha is 0 it will not be drawn at all)
        <sector_color> is the hex color code (see below) of the sector indicator
            (If alpha is 0 it will not be drawn at all)
        <drs_color> is the hex color code (see below) of the DRS start/end indicator
            (If alpha is 0 it will not be drawn at all)
        <drs_detection_color> is the hex color code (see below) of the DRS detection indicator
            (If alpha is 0 it will not be drawn at all)
"""

def log(text=""):
    print("%5i: %s" % (os.getpid(), text))

class Transformer:

    def __init__(self, scale, offsetx, offsety, rotation, aspect_ratio, centerx, centery):
        cr = math.cos(rotation)
        sr = math.sin(rotation)

        M1 = numpy.array(
            [[scale, 0, offsetx - centerx + 0.5],
             [0, scale, offsety - centery + 0.5],
             [0, 0, 1]])
        M2 = numpy.array(
            [[cr, -sr, 0],
             [sr,  cr, 0],
             [0 ,  0 , 1]])
        M3 = numpy.array(
            [[1, 0, centerx],
             [0, aspect_ratio, centery],
             [0, 0, 1]])
        self.M = numpy.dot(numpy.dot(M3, M2), M1)

    def apply(self, points):
        if len(points) == 0:
            return numpy.array([])

        points = numpy.array(points)
        x = points[:,0]
        y = points[:,2]
        p = numpy.zeros([len(x), 3])
        p[:,0] = x
        p[:,1] = y
        p[:,2] = 1
        return numpy.transpose(numpy.dot(self.M, numpy.transpose(p)))

def ac_path():
    script_file = os.path.realpath(__file__)
    script_dir = os.path.dirname(script_file)
    return os.path.join(script_dir, "..", "..", "..")

def resolve_script_relative_path(path):
    if getattr(sys, 'frozen', False):
        application_path = os.path.dirname(sys.executable)
    elif __file__:
        application_path = os.path.dirname(__file__)

    if os.path.isabs(path):
        return path
    else:
        path = os.path.join(application_path, path)
    return os.path.abspath(path)

def generate_png_file_fastlane(fai_file, map_file, ini_file, widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit, overrideOnlyIfNew):
    trackNodes = getNodesFromFastLane(fai_file)

    if len(trackNodes) == 0:
        raise Exception("fast_lane.ai file did not contain any valid nodes.")

    # main track points
    trackPoints = list(map(lambda node: node.position, trackNodes))

    # pitlane
    pitPoints = []
    pit_file = os.path.dirname(fai_file) + "/pit_lane.ai"
    if colors["pit"][3] > 0 and os.path.isfile(pit_file):
        pitPoints = list(map(lambda node: node.position, getNodesFromFastLane(pit_file)))

    # additional indicators
    additionalIndicators = []

    # sectors
    if colors["sector"][3] > 0:
        additionalIndicators += getSectorPoints(os.path.dirname(fai_file) + "/../data/overlays.ini", colors["sector"], trackPoints)
    # drs
    if colors["drs"][3] > 0 or colors["drs_detection"][3] > 0:
        additionalIndicators += getDrsPoints(os.path.dirname(fai_file) + "/../data/drs_zones.ini", colors["drs"], colors["drs_detection"], trackNodes)

    # reverse additional indicators, to draw latest find below first found
    return generate_png_file(trackPoints, pitPoints, additionalIndicators[::-1], map_file, ini_file, widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit, overrideOnlyIfNew)

def generate_png_file(rawTrackPoints, rawPitPoints, additionalIndicators, map_file, ini_file, widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit, overrideOnlyIfNew):
    # early break if map is already up to date
    if overrideOnlyIfNew and isMapUpToDate(ini_file, [widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit]):
        log("Track is already up to date. Remove '%s' to force re-generation." % (ini_file))
        return False

    # widths
    linew = widths["line"]
    outlinew = widths["outline"]
    mark_scale = widths["mark_scale"]
    mark_outline_width = widths["mark_outline"]
    linew_full = linew + 2*outlinew

    fill_color = colors["line"]
    outline_color = colors["outline"]
    pit_color = colors.get("pit", (128, 128, 128, 255))
    start_color = colors.get("start", (255, 0, 0, 255))

    # colors

    closed_track = False
    dist = numpy.linalg.norm(numpy.array(rawTrackPoints[0]) - numpy.array(rawTrackPoints[-1]))
    if dist < 100.:
        log("closed track detected. Point distance: %.1f" % dist)
        closed_track = True
        rawTrackPoints.append(rawTrackPoints[0])
    else:
        log("distance between last and first point is too large: %.1f" % dist)

    superSamplingFactor = 4

    spreadx, spready = 0, 0
    centerx, centery = 0, 0

    allPoints = rawTrackPoints + rawPitPoints

    if circle_fit == "circle":
        circleCenterX, circleCenterY, circleRadius = make_circle([(p[0], p[2]) for p in allPoints])

        spreadx = 2 * circleRadius
        spready = spreadx

        centerx = circleCenterX
        centery = circleCenterY
    else:
        minx = min(map(lambda p: p[0], allPoints))
        maxx = max(map(lambda p: p[0], allPoints))

        miny = min(map(lambda p: p[2], allPoints))
        maxy = max(map(lambda p: p[2], allPoints))

        # choose offset that avg(minx,maxx) and avg(miny,maxy) is mapped to center
        centerx = 0.5*(minx+maxx)
        centery = 0.5*(miny+maxy)

        if circle_fit == "centered_circle":
            circleRadius = max(map((lambda p: numpy.linalg.norm(numpy.array([p[0], p[2]]) - numpy.array([centerx, centery]))), allPoints))
            spreadx = 2 * circleRadius
            spready = spreadx
        else:
            spreadx = maxx-minx
            spready = maxy-miny

    if size_mode == "px":
        dimensions = calculateDimensionsForPx(png_size, adaptive_px_per_m, spreadx, spready, linew_full)
    elif size_mode == "m":
        dimensions = calculateDimensionsForMeter(png_size, adaptive_px_per_m, spreadx, spready, linew_full)
        linew, outlinew, mark_outline_width, linew_full = linew * dimensions[2], outlinew * dimensions[2], mark_outline_width * dimensions[2], linew_full * dimensions[2]
    else:
        log("Invalid size mode. Use -px or -m")
        sys.exit(1)

    width, height, scale = dimensions

    # scale*centerx + offsetx = width/2 -> offsetx = width/2 - scale*centerx
    offsetx = 0.5*width - scale*centerx
    offsety = 0.5*height - scale*centery

    t = Transformer(
        scale * superSamplingFactor,
        offsetx * superSamplingFactor,
        offsety * superSamplingFactor,
        0.0,
        1.0,
        0.5*width*superSamplingFactor,
        0.5*height*superSamplingFactor
    )
    p2d = t.apply(rawTrackPoints)
    pit2d = t.apply(rawPitPoints)

    # cast to int, because otherwise there are gaps at the joints
    points = list(map(lambda point: (int(point[0]), int(point[1])), list(p2d[:,0:2])))
    pitPoints = list(map(lambda point: (int(point[0]), int(point[1])), list(pit2d[:,0:2]))) if len(pit2d) > 0 else []

    drawWidths = (linew, outlinew, mark_scale, mark_outline_width, linew_full)
    drawColors = (fill_color, outline_color, pit_color, start_color)

    im = Image.new("RGBA", (width, height), (0,0,0,0))

    tileSize = 1024
    for x in range(0, width, tileSize):
        for y in range(0, height, tileSize):
            tile = drawTile([x, y, min(x+tileSize, width), min(y+tileSize, height)], superSamplingFactor, drawWidths, drawColors, points, pitPoints, additionalIndicators, closed_track, t, map_file)
            im.paste(tile, (x, y))

    log("Saving image ...")
    os.makedirs(os.path.dirname(map_file), exist_ok=True)
    im.save(map_file, "PNG")
    im.close()

    log("Saving ini file ...")
    os.makedirs(os.path.dirname(ini_file), exist_ok=True)
    save({'scale': scale,
          'offsetx': offsetx,
          'offsety': offsety,
          'map_size_x': width,
          'map_size_y': height,
          'circle_fit': circle_fit,
          'generation_hash': getGenerationHash(widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit)},
         ini_file)

    return True

def drawTile(boundaries, superSamplingFactor, widths, colors, pointsRaw, pitPointsRaw, additionalIndicators, closed_track, t, map_file):
    minX, minY, maxX, maxY = boundaries
    linew, outlinew, mark_scale, mark_outline_width, linew_full = widths
    fill_color, outline_color, pit_color, start_color = colors

    width, height = ((maxX - minX) * superSamplingFactor, (maxY - minY) * superSamplingFactor)

    centerOffset = int(linew_full*superSamplingFactor/2)

    points = list(map(lambda p: (p[0] - minX*superSamplingFactor, p[1] - minY*superSamplingFactor), pointsRaw))
    pitPoints = list(map(lambda p: (p[0] - minX*superSamplingFactor, p[1] - minY*superSamplingFactor), pitPointsRaw))

    im = Image.new("RGBA", (width, height), (0,0,0,0))
    draw = ImageDraw.Draw(im)
    # log("Drawing tile... (%i, %i)" % (minX, minY))

    if len(pitPoints) > 0:
        pitPointsVisibleList = mapToVisibleLines(width, height, pitPoints, linew_full*superSamplingFactor)
        for pitPointsVisible in pitPointsVisibleList:
            draw.line(pitPointsVisible, fill=(outline_color[:3] + (pit_color[3],)), width=int(0.5*linew_full*superSamplingFactor), joint='curve')
        for pitPointsVisible in pitPointsVisibleList:
            draw.line(pitPointsVisible, fill=pit_color, width=int(0.5*linew*superSamplingFactor), joint='curve')

    # draw start/end points as circles as the might be cropped on open ended lines
    draw.ellipse((points[0][0] - centerOffset, points[0][1] - centerOffset, points[0][0] + centerOffset, points[0][1] + centerOffset), fill=fill_color, outline=outline_color, width=int(outlinew*superSamplingFactor))
    draw.ellipse((points[-1][0] - centerOffset, points[-1][1] - centerOffset, points[-1][0] + centerOffset, points[-1][1] + centerOffset), fill=fill_color, outline=outline_color, width=int(outlinew*superSamplingFactor))


    pointsVisibleList = mapToVisibleLines(width, height, points, linew_full*superSamplingFactor)
    for pointsVisible in pointsVisibleList:
        draw.line(pointsVisible, fill=outline_color, width=int(linew_full*superSamplingFactor), joint='curve')
    for pointsVisible in pointsVisibleList:
        draw.line(pointsVisible, fill=fill_color, width=int(linew*superSamplingFactor), joint='curve')

    # generate start finish line
    if closed_track and start_color[3] > 0:
        startFinishPoints = numpy.array([points[-5],points[4]])
        startFinishCenter = numpy.mean(startFinishPoints, axis=0)

        direction = startFinishPoints[0] - startFinishPoints[1]

        drawCrossingLine(
            draw,
            startFinishCenter,
            startFinishPoints[0] - startFinishPoints[1],
            (2.5*linew*mark_scale + 2*mark_outline_width)*superSamplingFactor,
            (0.7*linew*mark_scale + 2*mark_outline_width)*superSamplingFactor,
            mark_outline_width*superSamplingFactor,
            start_color,
            outline_color
        )

    # draw additional indicators (sectors)
    for indicator in additionalIndicators:
        indicator2d = t.apply([indicator.coordinates])
        drawCrossingLine(
            draw,
            numpy.array([(int(indicator2d[0][0]) - minX*superSamplingFactor, int(indicator2d[0][1]) - minY*superSamplingFactor)]),
            numpy.array((indicator.orientation[0], indicator.orientation[2])),
            (2*linew*mark_scale + 2*mark_outline_width)*superSamplingFactor,
            (0.6*linew*mark_scale + 2*mark_outline_width)*superSamplingFactor,
            mark_outline_width*superSamplingFactor,
            indicator.color,
            outline_color
        )

    outFile = "%s_%i_%i.png" % (os.path.splitext(map_file)[0], minX, minY)
    finalIm = im.resize((int(width / superSamplingFactor), int(height / superSamplingFactor)), resample=Image.Resampling.LANCZOS)
    im.close()

    return finalIm

def mapToVisibleLines(width, height, points, linew_full):
    pointsList = []
    currentPoints = []

    for point in points:
        if (-2*linew_full < point[0] < width + 2*linew_full and
            -2*linew_full < point[1] < height + 2*linew_full):
            currentPoints.append(point)
        elif len(currentPoints) > 0:
            pointsList.append(currentPoints)
            currentPoints = []

    if len(currentPoints) > 0:
        pointsList.append(currentPoints)

    return pointsList

def drawCrossingLine(draw, point, circuitDirection, width, height, outlinew, fill_color, outline_color):
    dirLength = numpy.linalg.norm(circuitDirection)
    if dirLength == 0:
        return

    direction = circuitDirection * height/2 * (1/dirLength)
    orthDirection = numpy.array((-direction[1], direction[0])) * (width/height)

    crossingLine = numpy.array([
        point + orthDirection - direction,
        point - orthDirection - direction,
        point - orthDirection + direction,
        point + orthDirection + direction
    ])

    draw.polygon(list(crossingLine.flat), fill=fill_color, outline=outline_color, width=int(outlinew))


def save(register_info, out_file):
    scale = register_info['scale']
    ox = register_info['offsetx']
    oy = register_info['offsety']
    register_info['scale_inv'] = 1./float(scale)
    register_info['offsetx_scaled'] = float(ox)*1./float(scale)
    register_info['offsety_scaled'] = float(oy)*1./float(scale)

    open(out_file, 'w').write("""
[PARAMETERS]
WIDTH=%(map_size_x)f
HEIGHT=%(map_size_y)f
MARGIN=20
SCALE_FACTOR=%(scale_inv)f
X_OFFSET=%(offsetx_scaled)f
Z_OFFSET=%(offsety_scaled)f
DRAWING_SIZE=10
CIRCLE_FIT=%(circle_fit)s
GENERATION_HASH=%(generation_hash)s
""" % register_info )

def getGenerationHash(widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit):
    hexColors = {k: tuplecolor2hex(v) for k, v in colors.items()}
    parameters = {
        'widths': widths,
        'colors': hexColors,
        'png_size': png_size,
        'adaptive_px_per_m': adaptive_px_per_m,
        'size_mode': size_mode,
        'circle_fit': circle_fit
    }
    value = create_png_version + ":" + json.dumps(parameters, sort_keys=True)
    return hashlib.sha256(value.encode('utf-8')).hexdigest()

def hexcolor2tuple(cstr):
    assert(len(cstr) == 2*4+2)
    assert(cstr[0:2] == "0x")
    res = (
        int(cstr[2:4],base=16),
        int(cstr[4:6],base=16),
        int(cstr[6:8],base=16),
        int(cstr[8:10],base=16)
        )
    return res

def tuplecolor2hex(tup):
    return "%02x%02x%02x%02x" % tup

def calculateDimensionsForPx(png_size, adaptive_px_per_m, spreadx, spready, linew_full):
    aspectRatio = spreadx/spready # to automatically crop image

    scale = 1
    spaceAround = 3*linew_full

    if png_size > 0:
        if aspectRatio > 1:
            width = int(png_size)
            scale = (width - spaceAround) / spreadx
            height = int(math.ceil(spready * scale + spaceAround))
        else:
            height = png_size
            scale = (height - spaceAround) / spready
            width = int(math.ceil(spreadx * scale + spaceAround))
    elif adaptive_px_per_m > 0:
        width = int(math.ceil(spreadx * adaptive_px_per_m + spaceAround))
        height = int(math.ceil(spready * adaptive_px_per_m + spaceAround))
        scale = adaptive_px_per_m

    return (width, height, scale)

def calculateDimensionsForMeter(png_size, adaptive_px_per_m, spreadx, spready, linew_full):
    aspectRatio = spreadx/spready # to automatically crop image

    scale = 1
    spaceAround = 3*linew_full

    if png_size > 0:
        if aspectRatio > 1:
            width = int(png_size)
            scale = width / (spreadx + spaceAround)
            height = int(math.ceil((spready + spaceAround) * scale))
        else:
            height = png_size
            scale = height / (spready + spaceAround)
            width = int(math.ceil((spreadx + spaceAround) * scale))
    elif adaptive_px_per_m > 0:
        width = int(math.ceil((spreadx + spaceAround) * adaptive_px_per_m))
        height = int(math.ceil((spready + spaceAround) * adaptive_px_per_m))
        scale = adaptive_px_per_m

    return (width, height, scale)

# Data conventions: A point is a pair of floats (x, y). A circle is a triple of floats (center x, center y, radius).

# Returns the smallest circle that encloses all the given points. Runs in expected O(n) time, randomized.
# Input: A sequence of pairs of floats or ints, e.g. [(0,5), (3.1,-2.7)].
# Output: A triple of floats representing a circle.
# Note: If 0 points are given, None is returned. If 1 point is given, a circle of radius 0 is returned.
#
# Initially: No boundary points known
def make_circle(points):
    # Convert to float and randomize order
    shuffled = [(float(x), float(y)) for (x, y) in points]
    random.shuffle(shuffled)

    # Progressively add points to circle or recompute circle
    c = None
    for (i, p) in enumerate(shuffled):
        if c is None or not is_in_circle(c, p):
            c = _make_circle_one_point(shuffled[ : i + 1], p)
    return c

# One boundary point known
def _make_circle_one_point(points, p):
    c = (p[0], p[1], 0.0)
    for (i, q) in enumerate(points):
        if not is_in_circle(c, q):
            if c[2] == 0.0:
                c = make_diameter(p, q)
            else:
                c = _make_circle_two_points(points[ : i + 1], p, q)
    return c

# Two boundary points known
def _make_circle_two_points(points, p, q):
    circ = make_diameter(p, q)
    left  = None
    right = None
    px, py = p
    qx, qy = q

    # For each point not in the two-point circle
    for r in points:
        if is_in_circle(circ, r):
            continue

        # Form a circumcircle and classify it on left or right side
        cross = _cross_product(px, py, qx, qy, r[0], r[1])
        c = make_circumcircle(p, q, r)
        if c is None:
            continue
        elif cross > 0.0 and (left is None or _cross_product(px, py, qx, qy, c[0], c[1]) > _cross_product(px, py, qx, qy, left[0], left[1])):
            left = c
        elif cross < 0.0 and (right is None or _cross_product(px, py, qx, qy, c[0], c[1]) < _cross_product(px, py, qx, qy, right[0], right[1])):
            right = c

    # Select which circle to return
    if left is None and right is None:
        return circ
    elif left is None:
        return right
    elif right is None:
        return left
    else:
        return left if (left[2] <= right[2]) else right

def make_diameter(a, b):
    cx = (a[0] + b[0]) / 2
    cy = (a[1] + b[1]) / 2
    r0 = math.hypot(cx - a[0], cy - a[1])
    r1 = math.hypot(cx - b[0], cy - b[1])
    return (cx, cy, max(r0, r1))

def make_circumcircle(a, b, c):
    # Mathematical algorithm from Wikipedia: Circumscribed circle
    ox = (min(a[0], b[0], c[0]) + max(a[0], b[0], c[0])) / 2
    oy = (min(a[1], b[1], c[1]) + max(a[1], b[1], c[1])) / 2
    ax = a[0] - ox;  ay = a[1] - oy
    bx = b[0] - ox;  by = b[1] - oy
    cx = c[0] - ox;  cy = c[1] - oy
    d = (ax * (by - cy) + bx * (cy - ay) + cx * (ay - by)) * 2.0
    if d == 0.0:
        return None
    x = ox + ((ax*ax + ay*ay) * (by - cy) + (bx*bx + by*by) * (cy - ay) + (cx*cx + cy*cy) * (ay - by)) / d
    y = oy + ((ax*ax + ay*ay) * (cx - bx) + (bx*bx + by*by) * (ax - cx) + (cx*cx + cy*cy) * (bx - ax)) / d
    ra = math.hypot(x - a[0], y - a[1])
    rb = math.hypot(x - b[0], y - b[1])
    rc = math.hypot(x - c[0], y - c[1])
    return (x, y, max(ra, rb, rc))

_MULTIPLICATIVE_EPSILON = 1 + 1e-14

def is_in_circle(c, p):
    return c is not None and math.hypot(p[0] - c[0], p[1] - c[1]) <= c[2] * _MULTIPLICATIVE_EPSILON

# Returns twice the signed area of the triangle defined by (x0, y0), (x1, y1), (x2, y2).
def _cross_product(x0, y0, x1, y1, x2, y2):
    return (x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)

# fast_lane.ai helpers

def getSectorPoints(overlay_file, color, trackPoints):
    sectorPoints = []

    if os.path.isfile(overlay_file):
        overlaysConfig = configparser.ConfigParser(strict=False, inline_comment_prefixes=';')
        overlaysConfig.read(overlay_file, encoding='utf-8')

        for section in overlaysConfig.sections():
            if section.startswith("CHECKPOINT") and section != "CHECKPOINT_0":
                try:
                    coordinates = tuple(map(lambda val: float(val), overlaysConfig.get(section, "WORLD_POSITION").split(",")))
                    orientation = tuple(map(lambda val: float(val), overlaysConfig.get(section, "ORIENTATION").split(",")))

                    if checkIfPointIsOnTheTrack(trackPoints, coordinates):
                        sectorPoints.append(IndicatorLine(coordinates, orientation, color))
                except Exception as e:
                    log("Ignored sector entry: %s" % (e.message if hasattr(e, 'message') else e))

    return sectorPoints

def getDrsPoints(drs_file, color_drs, color_detection, trackNodes):
    drsPoints = []

    if os.path.isfile(drs_file):
        drsConfig = configparser.ConfigParser(strict=False, inline_comment_prefixes=';')
        drsConfig.read(drs_file, encoding='utf-8')

        lastPointDistance = numpy.linalg.norm(numpy.array(trackNodes[0].position) - numpy.array(trackNodes[-1].position))
        totalDistance = trackNodes[-1].distance + lastPointDistance

        for section in drsConfig.sections():
            if section.startswith("ZONE"):
                try:
                    for entry in drsConfig.options(section):
                        color = color_detection if entry.upper() == "DETECTION" else color_drs
                        if color[3] > 0:
                            distance = drsConfig.getfloat(section, entry)
                            (coordinates, direction) = getCoordinatesAndDirectionForTrackDistance(totalDistance, distance, trackNodes)
                            drsPoints.append(IndicatorLine(coordinates, direction, color))
                except Exception as e:
                    log("Ignored DRS entry: %s" % (e.message if hasattr(e, 'message') else e))

    return drsPoints

def getCoordinatesAndDirectionForTrackDistance(totalDistance, distance, trackNodes):
    index, node = next((idx, node) for (idx, node) in enumerate(trackNodes) if node.distance/totalDistance > distance)
    prevNode = trackNodes[index-1]

    positions = (numpy.array(prevNode.position), numpy.array(node.position))
    progresses = (prevNode.distance/totalDistance, node.distance/totalDistance)

    direction = positions[1] - positions[0]
    coordinates = positions[0] + direction * ((distance-progresses[0]) / (progresses[1] - progresses[0]))

    return (coordinates, direction)

def getNodesFromFastLane(fai_file):
    try:
        with open(fai_file, "rb") as buffer:
            log("Parsing fast_lane.ai: %s ..." % fai_file)
            header, length, lapTime, sampleCount = struct.unpack("4i", buffer.read(4 * 4))

            rawIdeal = []
            for i in range(length):
                rawIdeal.append(struct.unpack("4f i", buffer.read(4 * 5)))

            nodes = []
            extraCount = struct.unpack("i", buffer.read(4))[0]
            for i in range(extraCount):
                prevRawIdeal = rawIdeal[i - 1] if i > 0 else rawIdeal[length - 1]
                try:
                    node = TrackDetailNode(rawIdeal[i], prevRawIdeal, struct.unpack("18f", buffer.read(4 * 18)))
                    nodes.append(node)
                except:
                    # ignore invalid nodes and try to continue
                    pass

            return nodes
    except:
        return []

def checkIfPointIsOnTheTrack(trackPoints, point):
    for trackPoint in trackPoints:
        if numpy.linalg.norm(numpy.array(trackPoint) - numpy.array(point)) < 30.:
            return True
    return False

class TrackDetailNode:
    def __init__(self, rawIdeal, prevRawIdeal, rawDetail):
        self.id = rawIdeal[4]
        self.position = (rawIdeal[0], rawIdeal[1], rawIdeal[2])
        self.distance = rawIdeal[3]
        self.direction = -math.degrees(
            math.atan2(prevRawIdeal[2] - self.position[1],
                       self.position[0] - prevRawIdeal[0]))

        _wallLeft = rawDetail[5]
        _wallRight = rawDetail[6]

        self.wallLeft = slideVec2d(self.position, -self.direction + 90, _wallLeft)
        self.wallRight = slideVec2d(self.position, -self.direction - 90, _wallRight)
        self.trackCenter = numpy.mean(numpy.array([self.wallLeft, self.wallRight]), axis=0)

def slideVec2d(point, angle=0, offset=0):
    x = point[0] + math.cos(angle * math.pi / 180) * offset
    z = point[2] - math.sin(angle * math.pi / 180) * offset
    return (x, point[1], z)

class IndicatorLine:
    def __init__(self, coordinates, orientation, color):
        self.coordinates = coordinates
        self.orientation = orientation
        self.color = color

    def translateCopy(self, x, y, z):
        return IndicatorLine(
            (self.coordinates[0] + x, self.coordinates[1] + y, self.coordinates[2] + z),
            self.orientation,
            self.color
        )

# batch generation

def batchGenerateMaps(suffix, tracksDir, outDir, includedTracks, parameters):
    # iterate over selected/all tracks
    tracks = [d for d in os.listdir(tracksDir) if os.path.isdir(os.path.join(tracksDir, d))] if len(includedTracks) == 0 else [d for d in includedTracks if os.path.isdir(os.path.join(tracksDir, d))]

    trackArgs = list(map(lambda track: [track, parameters, suffix, tracksDir, outDir], tracks))

    pool = Pool(math.ceil(cpu_count() / 4))
    pool.map(generateTrackMap, trackArgs)

def generateTrackMap(args):
    try:
        trackName, parameters, suffix, tracksDir, outDir = args

        trackConfigName = trackName + suffix
        trackDir = os.path.join(tracksDir, trackName)

        # track without multiple layouts
        if os.path.isfile(os.path.join(trackDir, "ai/fast_lane.ai")):
            fai_file = os.path.join(trackDir, "ai/fast_lane.ai")
            map_file = os.path.join(outDir, trackConfigName + ".png")
            ini_file = os.path.join(outDir, trackConfigName + ".ini")
            try:
                generate_png_file_fastlane(fai_file, map_file, ini_file, *parameters)
            except Exception as e:
                log("Ignored %s as the fast_lane.ai file could not be parsed: %s" % (trackConfigName, e.message if hasattr(e, 'message') else e))
        # track with multiple layouts (or no fast_lane.ai at all)
        else:
            layouts = [d for d in os.listdir(trackDir) if os.path.isdir(os.path.join(trackDir, d)) and os.path.isfile(os.path.join(os.path.join(trackDir, d), "ai/fast_lane.ai"))]
            for layout in layouts:
                trackConfigName = trackName + "/" + layout + suffix
                layoutDir = os.path.join(trackDir, layout)
                fai_file = os.path.join(layoutDir, "ai/fast_lane.ai")
                map_file = os.path.join(outDir, trackConfigName + ".png")
                ini_file = os.path.join(outDir, trackConfigName + ".ini")
                try:
                    generate_png_file_fastlane(fai_file, map_file, ini_file, *parameters)
                except Exception as e:
                    log("Ignored %s as the fast_lane.ai file could not be parsed: %s" % (trackConfigName, e.message if hasattr(e, 'message') else e))
    except:
        log("Unexpected error for: %s" % args)

def isMapUpToDate(ini_file, parameters):
    mapConfig = configparser.ConfigParser(strict=False)
    mapConfig.read(ini_file, encoding='utf-8')
    return mapConfig.get("PARAMETERS", "GENERATION_HASH", fallback="") == getGenerationHash(*parameters)

def parseGenerationParametersFromArgs(args):
    widths = {}
    colors = {}
    widths["line"]  = float(args.pop(0))
    widths["outline"] = float(args.pop(0))
    colors["outline"] = hexcolor2tuple(args.pop(0))
    colors["line"] = hexcolor2tuple(args.pop(0))
    colors["pit"] = hexcolor2tuple(args.pop(0))
    widths["mark_scale"] = float(args.pop(0)) if len(args) > 0 else 1
    widths["mark_outline"] = float(args.pop(0)) if len(args) > 0 else 1
    colors["start"] = hexcolor2tuple(args.pop(0)) if len(args) > 0 else (255, 0, 0, 255)
    colors["sector"] = hexcolor2tuple(args.pop(0)) if len(args) > 0 else (255, 255, 0, 255)
    colors["drs"] = hexcolor2tuple(args.pop(0)) if len(args) > 0 else (0, 128, 255, 255)
    colors["drs_detection"] = hexcolor2tuple(args.pop(0)) if len(args) > 0 else (0, 200, 0, 255)
    return [widths, colors]

def formatHex(baseHex):
    cleanHex = baseHex.replace('#', '')
    for idx in range(8 - len(cleanHex)):
        cleanHex += "f"
    return "0x" + cleanHex

# main

if __name__ == "__main__":
    # On Windows calling this function is necessary (with multiprocessing import)
    freeze_support()

    png_size = 512
    adaptive_px_per_m = 0
    size_mode = "px"
    included_tracks = []
    circle_fit = ""
    overrideOnlyIfNew = True

    args = sys.argv[1:]

    while len(args) > 0:
        arg = args.pop(0)

        if arg == "-s":
            png_size = int(args.pop(0))
            adaptive_px_per_m = 0
        elif arg == "-a":
            adaptive_px_per_m = float(args.pop(0))
            png_size = 0
        elif arg == "-m":
            size_mode = "m"
        elif arg == "-px":
            size_mode = "px"
        elif arg == "-c":
            circle_fit = "circle"
        elif arg == "-cc":
            circle_fit = "centered_circle"
        elif arg == "-f":
            overrideOnlyIfNew = False
        # parse args for fast_lane.ai
        # Test command circuit: py apps/lua/map_display/create_png.py -s 2048 -fai /e/Games/Steam/steamapps/common/assettocorsa/content/tracks/ks_nurburgring/layout_gp_b/ai/fast_lane.ai ./test_file 30 3 0x000000ff 0xffffffff 0x777777ff 1 1 0xff0000ff 0xFFFF00FF 0x0080FFFF 0x00BF0000
        # Test command line: py apps/lua/map_display/create_png.py -s 2048 -fai /e/Games/Steam/steamapps/common/assettocorsa/content/tracks/pacific_coast_hwy/northbound/ai/fast_lane.ai ./test_file 30 3 0x000000ff 0xffffffff 0x777777ff 1 1 0xff0000ff 0xFFFF00FF 0x0080FFFF 0x00BF0000
        elif arg == "-fai" and len(sys.argv) > 7 :
            fai_file = args.pop(0)
            if '*' in fai_file:
                fai_files = glob.glob(fai_file)
            else:
                fai_files = [fai_file]
            for fai_file in fai_files:
                file_path = args.pop(0)
                map_file = file_path + ".png"
                ini_file = file_path + ".ini"
                widths, colors = parseGenerationParametersFromArgs(args)
                result = generate_png_file_fastlane(fai_file, map_file, ini_file, widths, colors, png_size, adaptive_px_per_m, size_mode, circle_fit, overrideOnlyIfNew)

                if not result:
                    sys.exit(10) # exit with code 10 if no map was generated
        # batch generation
        # Test command full: py apps/lua/map_display/create_png.py -s 1024 -px -t "ks_nurburgring" -bf /e/Games/Steam/steamapps/common/assettocorsa/content/tracks ../../../test_out 18 2 0x000000ff 0xffffffff 0x777777ff 1 1 0xff0000ff 0xFFFF00FF 0x0080FFFF 0x00BF0000
        # Test command mini: py apps/lua/map_display/create_png.py -a 0.5 -m -t "ks_nurburgring" -bm /e/Games/Steam/steamapps/common/assettocorsa/content/tracks ../../../test_out 32 3 0x000000ff 0xffffffff 0x777777ff 1 2 0xff0000ff 0xFFFF00FF 0x0080FFFF 0x00BF0000
        elif arg in ["-bf", "-bm"] and len(sys.argv) > 7:
            tracksDir = resolve_script_relative_path(args.pop(0))
            outDir = resolve_script_relative_path(args.pop(0))
            widths, colors = parseGenerationParametersFromArgs(args)

            if arg == "-bf":
                batchGenerateMaps("_full", tracksDir, outDir, includedTracks, [widths, colors, png_size, 0, "px", circle_fit, overrideOnlyIfNew])
            if arg == "-bm":
                batchGenerateMaps("_mini", tracksDir, outDir, includedTracks, [widths, colors, 0, adaptive_px_per_m, "m", circle_fit, overrideOnlyIfNew])
        elif arg == "-t":
            includedTracksString = args.pop(0)
            includedTracks = includedTracksString.split(",") if len(includedTracksString) > 0 else []
        else:
            log(usage)
            sys.exit(1)
