MapGenerator = class('MapGenerator')

function MapGenerator:initialize(suffix, mode)
    self.suffix = suffix
    self.mode = mode
    self.generationRunning = false
end

---@param config table<string>
---@param callback fun(fileImage?: string, fileIni?: string)
function MapGenerator:generatePNG(config, forceGeneration, callback)
    local fullTrackPath = ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\')

    local outPath = self:getTrackMapPath()

    ac.debug('MapGenerator' .. self.suffix .. '_fullTrackPath', fullTrackPath)
    ac.debug('MapGenerator' .. self.suffix .. '_outputPath', outPath)

    if self.generationRunning then
        return
    end

    local exe = APP_DIR .. 'create_png.exe'
    if io.exists(exe) then
        self.generationRunning = true

        local arguments = self:createGenerationArguments(config, fullTrackPath, outPath, forceGeneration, false)

        ac.debug('MapGenerator' .. self.suffix .. '_arguments', table.join(arguments, ' '))

        os.runConsoleProcess(
            {
                filename = exe,
                arguments = {
                    ' ', -- first argument is ignored as it's usually the file reference
                    table.unpack(arguments),
                },
                rawArguments = true,
                workingDirectory = '',
                timeout = 15000,
                terminateWithScript = true
            },
            function(err, data)
                self.generationRunning = false

                ac.debug('MapGenerator' .. self.suffix .. '_exitCode', data.exitCode)
                ac.debug('MapGenerator' .. self.suffix .. '_stdout', data.stdout)
                ac.debug('MapGenerator' .. self.suffix .. '_stderr', data.stderr)

                if data.exitCode == 0 or data.exitCode == 10 then
                    self.generationFailed = false
                    callback(outPath .. '.png', outPath .. '.ini')
                else
                    self.generationFailed = true
                    callback()
                end
            end)
    else
        self.generationFailed = true
        callback()
    end
end

function MapGenerator:createBatchScript(config)
    local filePath = self:getBatchScriptPath()

    io.save(filePath, string.format(([[
        set /P tracks="Enter comma separated track names (no spaces, empty for ALL): "
        .\create_png.exe -t "%%tracks%%" %s
        pause
    ]]):gsub('^ +', ''):gsub('\n +', '\n'),
        table.join(self:createGenerationArguments(config, '..\\..\\..\\content\\tracks', 'data', false, true), ' ')))

    return filePath:sub(APP_DIR:len() + 1)
end

function MapGenerator:getBatchScriptPath()
    local suffix = self.mode == MapGenerator.SizeMode.Adaptive and 'mini_maps' or 'full_maps'
    return APP_DIR .. 'generate_' .. suffix .. '.bat'
end

function MapGenerator:createGenerationArguments(config, trackPath, outPath, forceGeneration, isBatch)
    local batchMode = self.mode == MapGenerator.SizeMode.Adaptive
        and '-bm'
        or '-bf'

    local sizeParams = self.mode == MapGenerator.SizeMode.Adaptive
        and string.format("-a %.1f -m", config.pxPerM)
        or string.format("-s %.0f -px", config.imageSize) -- TODO add circle fit option

    local shapeParams = ""
    if self.mode == MapGenerator.SizeMode.Static then
        shapeParams = (config.shapeMode == MapGenShape.Circle and "-c")
            or (config.shapeMode == MapGenShape.CircleCentered and "-cc")
            or ""
    end

    local modeParams = isBatch
        and string.format('%s "%s" "%s"', batchMode, trackPath, outPath)
        or string.format('-fai "%s" "%s"', trackPath .. '\\ai\\fast_lane.ai', outPath)

    return {
        forceGeneration and '-f' or '',
        sizeParams,
        shapeParams,
        modeParams,
        config.lineWidth,
        config.outlineWidth,
        '0x' .. RgbmToHex(config.outlineColor),
        '0x' .. RgbmToHex(config.lineColor),
        '0x' .. RgbmToHex(config.pitLineColor),
        config.marksScaling,
        config.marksOutlineWidth,
        '0x' .. RgbmToHex(config.startColor),
        '0x' .. RgbmToHex(config.sectorColor),
        '0x' .. RgbmToHex(config.drsColor),
        '0x' .. RgbmToHex(config.drsDetection),
    }
end

function MapGenerator:isGenerationInProgress()
    return self.generationRunning
end

function MapGenerator:isGenerationSupported()
    return io.exists(ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\') .. '\\ai\\fast_lane.ai')
        and not self.generationFailed
end

function MapGenerator:getTrackMapPath()
    return APP_DIR .. 'data\\' .. ac.getTrackFullID('\\') .. self.suffix
end

MapGenerator.SizeMode = {
    Static = 0,
    Adaptive = 1
}
