require('helpers/model/driver')
local json = require('json')

Car = class('Car')

function Car:initialize(id) -- constructor
  self.id = id
  self.carTechnicalID = ac.getCarID(id)
  self.carName = ac.getCarName(id, true)
  self.isActive = false
  self.isHidingLabels = false
  self.racePosition = id
  self.worldPosition = vec3()
  self.lookVector = vec3()
  self.speed = 0
  self.inPitLane = false
  self.lapCount = 0
  self.lapProgress = 0
  self.totalLaps = 0
  self.bestLap = 999999999
  self.lastFrameLapTime = -1
  self.carTags = self:parseCarTags()
  self.driver = Driver(id)
  self.crossedStart = false
  self.hasStartedLap = false
end

function Car:update(dt, index, sim)
  local carState = ac.getCar(self.id)

  if not carState then
    return
  end

  self.isActive = carState.isActive
  self.isHidingLabels = carState.isHidingLabels

  if self.isActive then
    self.racePosition = index
    self.worldPosition = carState.position
    self.lookVector = carState.look
    self.speed = carState.speedKmh
    self.inPitLane = carState.isInPitlane

    self.lapProgress = carState.splinePosition

    -- lap count correction for race start
    if not self.crossedStart and self.lapProgress < 0.5 then
      self.crossedStart = true
    elseif not sim.isSessionStarted and self.crossedStart and self.lapProgress > 0.5 then
      self.crossedStart = false
    end

    -- lap count, making sure it's only updated in sync with the lapProgress
    if self.lapCount < carState.sessionLapCount or self.lapProgress > 0.5 then
      self.lapCount = self.crossedStart
          and carState.sessionLapCount
          or (carState.sessionLapCount - 1)
    end
    self.totalLaps = self.lapCount + self.lapProgress

    -- workaround for online because lastLapTimeMs and bestLapTimeMs are not
    -- working in online sessions yet. So we need to prevent only half witnessed laps to count as unbeatable times
    if sim.isOnlineRace then
      if self.inPitLane then
        -- AC accepts laps as valid that are started from the pits
        self.hasStartedLap = true
      elseif carState.lapTimeMs < self.lastFrameLapTime then
        if not self.hasStartedLap then
          self.hasStartedLap = true
        elseif carState.isLastLapValid then
          self.bestLap = self.lastFrameLapTime + dt
        end
      end
      self.lastFrameLapTime = carState.lapTimeMs
    elseif carState.bestLapTimeMs > 0
        and carState.bestLapTimeMs < self.bestLap then
      self.bestLap = carState.bestLapTimeMs
    end
    self.driver:update(dt)
  end
end

function Car:parseCarTags()
  local class = try(function()
    local file = ac.getFolder(ac.FolderID.ContentCars) .. '\\' .. ac.getCarID(self.id) .. '\\ui\\ui_car.json'
    local jsonString = ''
    for line in io.lines(file) do
      --\uFEFF
      jsonString = jsonString .. line:gsub('[\xEF\xBB\xBF]', '')
    end
    local carData = json:decode(jsonString)

    return carData ~= nil and carData.tags or {}
  end, function(err)
    ac.debug('classTagError_' .. ac.getCarID(self.id), err)
  end)

  return class or ac.getCarTags(self.id) or {}
end

function Car:activeStateComparison(otherCar)
  if self.inPitLane ~= otherCar.inPitLane then
    return self.inPitLane and 1 or -1
  end
  return 0
end

function Car:compareDistance(otherCar)
  local diff = otherCar.totalLaps - self.totalLaps
  if diff == 0 then
    self:compareRacePosition(otherCar)
  end
  return diff
end

function Car:compareTime(otherCar)
  local diff = self.bestLap - otherCar.bestLap
  if diff == 0 then
    diff = self:compareRacePosition(otherCar)
  end
  return diff
end

function Car:compareRacePosition(otherCar)
  local diff = self.racePosition - otherCar.racePosition
  if diff == 0 then
    diff = self.id - otherCar.id
  end
  return diff
end

function Car.getComparatorForSession()
  local isRace = ac.getSim().raceSessionType == ac.SessionType.Race
  local func = isRace and Car.compareDistance or Car.compareTime
  return function(a, b)
    if a.isActive ~= b.isActive then
      return a.isActive
    end
    return func(a, b) < 0
  end
end
