require('widgets/map_progress')
require('widgets/map_full_settings')

local zeroVector = vec2()

MapFull = class('MapFull', MapProgress)
function MapFull:initialize(name) -- constructor
  ---@type nil|table
  self.teleportPoints = nil

  MapProgress.initialize(self, name)
end

function MapFull:initializeSettings()
  self.settings = MapFullSettings(self.name)
end

function MapFull:draw()
  self.visible = true

  local drawOffset = ui.getCursor()
  local contentSize = ui.windowSize():sub(drawOffset)
  local contentCenter = drawOffset:clone():add(contentSize / 2)
  local maxSpace = math.min(contentSize.x, contentSize.y)
  local imageSize = ui.imageSize(self.settings.mapInfo.ref)

  local imageScale = self:getScale(maxSpace, imageSize)

  -- early exit if image not ready
  if not imageScale then
    local offset = vec2(20, 20)
    ui.drawLoadingSpinner(contentCenter - offset, contentCenter + offset)
    return
  end

  local scaledImageSize = imageSize * imageScale
  local scaledImageCenter = scaledImageSize / 2

  local rotation = self:getImageRotation()

  local positionOffset = self:getImagePositionOffset(imageScale, scaledImageCenter, contentSize, rotation)
  local imageOffset = contentSize:clone()
      :sub(scaledImageSize)
      :scale(0.5)
      :add(drawOffset)
      :add(positionOffset)

  -- start to clip content
  local clipP1, clipP2 = self:getContentClip(maxSpace, contentCenter)
  self:startContentClip(clipP1, clipP2)

  ui.beginRotation()
  ui.drawImage(
    self.settings.mapInfo.ref,
    imageOffset,
    imageOffset + scaledImageSize,
    self:calculateMapColor())
  ui.endPivotRotation(90 - rotation, imageOffset + scaledImageCenter)

  self:drawArrows({
    imageScale = imageScale,
    scaledImageCenter = scaledImageCenter,
    imageOffset = imageOffset,
    rotation = rotation
  })

  self:drawTeleports(imageScale, scaledImageCenter, imageOffset, rotation)

  -- end content clip
  self:endContentClip(clipP1, clipP2)

  -- debug only
  if self.settings.store.drawDebugLines then
    ui.beginRotation()
    -- debug rect for map image boundaries
    local imageMaxSize = math.max(imageSize.x, imageSize.y)
    local scaledImageBoundaries = vec2(imageMaxSize, imageMaxSize):scale(imageScale)
    ui.drawRect(
      (contentSize - scaledImageBoundaries):scale(1 / 2):add(drawOffset):add(positionOffset),
      (contentSize + scaledImageBoundaries):scale(1 / 2):add(drawOffset):add(positionOffset),
      rgbm(1, 0, 0, 1))
    ui.endPivotRotation(90 - rotation, imageOffset + scaledImageCenter)

    -- debug circle for rotated image boundaries
    ui.drawCircle(contentCenter, (maxSpace - self:getImageMargin()) / 2, rgbm(0, 0, 1, 1), 64)

    -- debug rect for drawable content size
    ui.drawRect(
      ui.getCursor(),
      ui.windowSize(),
      rgbm(1, 1, 0, 1))
  end
end

function MapFull:getScale(maxSpace, imageSize)
  local imageMaxSize = math.max(imageSize.x, imageSize.y)
  local imageRadius = (self.settings.mapInfo.circleFit or self.settings.full.marginNoRotation)
      and imageMaxSize
      or math.sqrt(2 * (imageMaxSize * imageMaxSize))

  return imageRadius > 0 and (maxSpace - self:getImageMargin()) / imageRadius or nil
end

function MapFull:getImageMargin()
  return self.settings.full.marginStatic
end

function MapFull:getImageRotation()
  return self.settings.trackStore.rotation
end

function MapFull:getImagePositionOffset(imageScale, scaledImageCenter, contentSize, rotation)
  return zeroVector
end

function MapFull:getContentClip(maxSpace, contentCenter)
  local halfSizeOverlay = vec2(maxSpace, maxSpace):scale(0.5)
  return contentCenter:clone():sub(halfSizeOverlay), contentCenter:clone():add(halfSizeOverlay)
end

function MapFull:startContentClip(p1, p2)
  -- do nothing
end

function MapFull:endContentClip(p1, p2)
  -- do nothing
end

function MapFull:drawTeleports(imageScale, imageCenter, imageOffset, rotation)
  if self.teleportPoints == nil or not ui.windowHovered() then
    return
  end

  table.forEach(self.teleportPoints, function(item, idx)
    local circleRadius = self.settings.freeroamStore.teleportCircleRadius
    local circleSteps = 24 -- roughly every 10px
    local pointOnMap = self:transformPointToMap(vec2(item.location.x, item.location.z), imageScale, imageCenter, rotation)
    -- visible circle
    ui.drawCircleFilled(pointOnMap:add(imageOffset), circleRadius, self.settings.freeroamStore.teleportCircleColor,
      circleSteps)
    ui.drawCircle(pointOnMap, circleRadius, self.settings.freeroamStore.teleportCircleOutline, circleSteps, 2)
    -- click items
    ui.setCursor(pointOnMap:sub(circleRadius))
    ui.dummy(vec2(circleRadius * 2, circleRadius * 2))
    if ui.itemHovered() then
      ui.tooltip(function()
        local text = table.join(table.filter({ item.group, item.name }, function(item) return item end), '/') or ''
        ui.text(text)
      end)
    end
    if ui.itemClicked() then
      ac.teleportToServerPoint(idx - 1)
    end
  end)
end

function MapFull:calculateCirclePositionOnMap(car, posCalculationParameters)
  local pos = car.worldPosition
  return self:transformPointToMap(
        vec2(pos.x, pos.z),
        posCalculationParameters.imageScale,
        posCalculationParameters.scaledImageCenter,
        posCalculationParameters.rotation)
      :add(posCalculationParameters.imageOffset)
end

function MapFull:calculateCircleScaledDirectionOnMap(car, posCalculationParameters, circleRadius)
  local lookVector = car.lookVector
  return RotatePoint(
        vec2(lookVector.x, lookVector.z):normalize(),
        zeroVector,
        posCalculationParameters.rotation)
      :scale(circleRadius)
end

function MapFull:transformPointToMap(point, imageScale, imageCenter, rotation)
  point
      :add(self.settings.mapInfo.offset)
      :scale(imageScale / self.settings.mapInfo.scaleFactor)

  return RotatePoint(
    point,
    imageCenter,
    rotation)
end

function MapFull:lazyUpdate()
  MapProgress.lazyUpdate(self)

  self:initOnlineTeleportPoints()
end

function MapFull:initOnlineTeleportPoints()
  if self.settings:attributeChanged('enableTeleports') then
    self.teleportPoints = nil
  end

  if self.teleportPoints == nil and self.settings.freeroamStore.enableTeleports and ac.getSim().isOnlineRace then
    local onlineExtras = ac.INIConfig.onlineExtras()
    if onlineExtras ~= nil then
      self.teleportPoints = self:parseTeleportPoints(onlineExtras)
    end
  elseif not ac.getSim().isOnlineRace then
    self.teleportPoints = {}
  end
end

function MapFull:parseTeleportPoints(ini)
  local teleports = {}
  for a, b in ini:iterateValues('TELEPORT_DESTINATIONS', 'POINT') do
    local n = tonumber(b:match('%d+')) + 1

    if teleports[n] == nil then
      for i = #teleports, n do
        if teleports[i] == nil then teleports[i] = {} end
      end
    end

    local suffix = b:match('_(%a+)$')
    if suffix == nil then
      teleports[n].name = ini:get('TELEPORT_DESTINATIONS', b, 'noname' .. n - 1)
    elseif suffix == 'POS' then
      teleports[n].location = ini:get('TELEPORT_DESTINATIONS', b, vec3())
    elseif suffix == 'HEADING' then
      teleports[n].direction = ini:get('TELEPORT_DESTINATIONS', b, 0)
    elseif suffix == 'GROUP' then
      teleports[n].group = ini:get('TELEPORT_DESTINATIONS', b, 'group')
    end
    teleports[n].index = n
  end

  local filteredTeleports = table.filter(teleports, function(item) return item ~= nil and item.name ~= nil end)
  table.sort(filteredTeleports,
    function(a, b)
      return a.index < b.index
    end)
  return filteredTeleports
end
