require('widgets/map_progress_settings')
require('helpers/map_generation')
require('widgets/settings/map_driver_settings')
require('widgets/settings/map_class_settings')

MapFullSettings = class('MapFullSettings', MapProgressSettings)

MapFullSettings.defaultGeneral = require('widgets/settings/general_defaults')
MapFullSettings.defaultFull = require('widgets/settings/full_defaults')
MapFullSettings.defaultMapGen = require('widgets/settings/map_gen_defaults')
MapFullSettings.defaultTrack = require('widgets/settings/track_defaults')
MapFullSettings.defaultFreeroam = require('widgets/settings/freeroam_defaults')

function MapFullSettings:initialize(name) -- constructor
  MapProgressSettings.initialize(self, name)

  self.full = ac.storage(self.defaultFull(), name .. '_Full')

  self.trackStore = ac.storage(self.defaultTrack(), name .. '_' .. ac.getTrackFullID('\\'))

  self.mapInfo = {
    ref = '',
    scaleFactor = 1,
    offset = vec2()
  }
  self:initializeMapGenerator()

  self:initMapImage()
end

function MapFullSettings:initializeMapGenerator()
  self.mapGenStore = ac.storage(self.defaultMapGen(), self.name .. '_MapGeneration')
  self.mapGenerator = MapGenerator('_full', MapGenerator.SizeMode.Static)
end

function MapFullSettings:mainContent()
  ui.tabBar('tabs', ui.TabBarFlags.FittingPolicyScroll + ui.TabBarFlags.TabListPopupButton, function()
    ui.tabItem('General', Bind(self.tabMain, self))
    ui.tabItem('Full Map', Bind(self.tabFull, self))
    ui.tabItem('Track', Bind(self.tabTrack, self))
    ui.tabItem('Drivers', Bind(self.tabDrivers, self))
    ui.tabItem('Map Generation', Bind(self.tabMapGen, self))
    ui.tabItem('Classes', Bind(self.tabClasses, self))
    ui.tabItem('Freeroam', Bind(self.tabFreeroam, self))
    ui.tabItem('Debug', Bind(self.tabDebug, self))
    return nil
  end)
end

function MapFullSettings:tabFull()
  ui.text("Full Map")
  SpacingY()

  self.full.marginStatic = TooltipSlider('Margin', self.full.marginStatic, -1000, 1000, '%.0f px')

  if ui.checkbox('Enable Rotation Margin', not self.full.marginNoRotation) then
    self.full.marginNoRotation = not self.full.marginNoRotation
  end
  if self.full.marginNoRotation then
    ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.yellow)
    ui.textWrapped(
      "When disabling the rotation margin, also check the config Map Generation > Shape Optimization. 'Square' uses the available space more efficiently if the map is never rotated.",
      ui.availableSpaceX())
    ui.popStyleColor()
  end

  SpacingSeparator()

  -- Reset
  if ui.button('Reset to default') then
    table.forEach(self.defaultFull(), function(item, key)
      self.full[key] = nil
    end)
  end
end

function MapFullSettings:tabTrack()
  ui.text('Configuration for: ' .. ac.getTrackName())
  SpacingY()

  self:tabTrackSelection()
  self:tabTrackRotation()

  SpacingSeparator()

  -- Reset
  if ui.button('Reset to default') then
    table.forEach(self.defaultTrack(), function(item, key)
      self.trackStore[key] = nil
    end)
    self:initMapImage()
  end
end

function MapFullSettings:tabTrackSelection()
  ui.combo('##Selected Map File', 'Select Map Image', function()
    local generatedImage = self.mapGenerator:getTrackMapPath() .. '.png'
    if io.fileExists(generatedImage) then
      if ui.selectable('Generated Map', self.trackStore.fileImage == generatedImage) then
        self:setFile(generatedImage, self.mapGenerator:getTrackMapPath() .. '.ini')
      end
    end
    -- other maps
    local maps = io.scanDir(ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\'), '\\map*.png')
    table.forEach(maps, function(item)
      local fullPath = ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\') .. '\\' .. item
      if ui.selectable(item, self.trackStore.fileImage == fullPath) then
        self:setFile(
          fullPath,
          ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\') .. '\\data\\map.ini')
      end
    end)
  end)
  if ui.itemHovered() then
    ui.tooltip(function()
      ui.text("If 'Generated Map' is missing, run 'Update' in tab 'Map Generation' first")
      ui.text("If it's still missing afterwards, track generation is not supported for the current track")
    end)
  end
end

function MapFullSettings:tabTrackRotation()
  self.trackStore.rotation = TooltipSlider('Rotation', self.trackStore.rotation, 0, 359, '%.0f°')
  if self.trackStore.rotation ~= 0 and self.full.marginNoRotation then
    ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.orange)
    ui.textWrapped(
      "Currently the rotation margin is disabled in Full Map settings. Make sure to enable that margin so that the map is not cut off.",
      ui.availableSpaceX())
    ui.popStyleColor()
  end
end

function MapFullSettings:tabMapGen()
  ui.text("Map Generation")
  SpacingY()

  if not self.mapGenerator:isGenerationSupported() then
    ui.textColored('Map generation failed.', rgbm.colors.red)
    ui.text('This could have multiple reasons:')
    ui.bullet()
    ui.textWrapped('This track does not support Map Generation as the AI line is missing or invalid (fast_lane.ai).',
      ui.availableSpaceX())
    ui.bullet()
    ui.textWrapped('The create_png.exe file is missing in the app directory.',
      ui.availableSpaceX())
    SpacingSeparator()
    -- make rest of this tab gray
    ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.gray)
  end

  -- general
  if ui.checkbox('Enable automatic map generation', self.mapGenStore.enableMapGeneration) then
    self.mapGenStore.enableMapGeneration = not self.mapGenStore.enableMapGeneration
  end

  if ui.checkbox('Auto select generated map', self.mapGenStore.selectGeneratedMap) then
    self.mapGenStore.selectGeneratedMap = not self.mapGenStore.selectGeneratedMap
  end

  SpacingSeparator()

  -- widths
  self:tabMapGenSizeConfig()
  self.mapGenStore.lineWidth = TooltipSlider('Line Width', self.mapGenStore.lineWidth, 1, 100, '%.0f px')
  self.mapGenStore.outlineWidth = TooltipSlider('Outline Width', self.mapGenStore.outlineWidth, 0, 100, '%.0f px')
  self.mapGenStore.marksScaling = TooltipSlider(
    'Marks Scaling', self.mapGenStore.marksScaling, 0.5, 2.0, '%.1fx')
  self.mapGenStore.marksOutlineWidth = TooltipSlider(
    'Marks Outline Width', self.mapGenStore.marksOutlineWidth, 0, 100, '%.0f px')

  SpacingSeparator()

  -- colors
  ui.columns(2, false)

  if ui.colorButton(
        'Line Color##',
        self.mapGenStore.lineColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.lineColor = self.mapGenStore.lineColor
  end
  ui.sameLine()
  ui.text('Line Color')

  ui.nextColumn()

  if ui.colorButton(
        'Outline Color##',
        self.mapGenStore.outlineColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.outlineColor = self.mapGenStore.outlineColor
  end
  ui.sameLine()
  ui.text('Outline Color')

  ui.nextColumn()

  if ui.colorButton(
        'Pitline Color##',
        self.mapGenStore.pitLineColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.pitLineColor = self.mapGenStore.pitLineColor
  end
  ui.sameLine()
  ui.text('Pitline Color')

  ui.nextColumn()

  if ui.colorButton(
        'Start/Finish Color##',
        self.mapGenStore.startColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.startColor = self.mapGenStore.startColor
  end
  ui.sameLine()
  ui.text('Start/Finish Color')

  ui.nextColumn()

  if ui.colorButton(
        'Sector Color##',
        self.mapGenStore.sectorColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.sectorColor = self.mapGenStore.sectorColor
  end
  ui.sameLine()
  ui.text('Sector Color')

  ui.nextColumn()

  if ui.colorButton(
        'DRS Color##',
        self.mapGenStore.drsColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.drsColor = self.mapGenStore.drsColor
  end
  ui.sameLine()
  ui.text('DRS Color')

  ui.nextColumn()

  if ui.colorButton(
        'DRS Detection Color##',
        self.mapGenStore.drsDetection,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.mapGenStore.drsDetection = self.mapGenStore.drsDetection
  end
  ui.sameLine()
  ui.text('DRS Detection Color')

  ui.nextColumn()
  ui.nextColumn()

  SpacingY()

  CopyPaste('mapGenColors', 'Colors', self.mapGenStore,
    { 'lineColor', 'outlineColor', 'pitLineColor', 'startColor', 'sectorColor', 'drsColor', 'drsDetection' })

  ui.columns()

  -- remove gray color for this tab
  if not self.mapGenerator:isGenerationSupported() then
    ui.popStyleColor()
  end

  self:tabMapGenBatchGeneration()

  SpacingSeparator()

  -- Regenerate
  if ui.button('Update') then
    self:generateMap(true, true)
  end
  if ui.itemHovered() then
    ui.tooltip(function()
      ui.text("Regenerates image, might take some time")
    end)
  end

  ui.sameLine()

  if self.mapGenerator:isGenerationInProgress() then
    ui.drawLoadingSpinner(ui.getCursor(),
      ui.getCursor():add(vec2(ui.textLineHeightWithSpacing(), ui.textLineHeightWithSpacing())))

    ui.sameLine(0, ui.textLineHeightWithSpacing() * 2)
  end

  -- Reset
  if ui.button('Reset to default') then
    table.forEach(self.defaultMapGen(), function(item, key)
      self.mapGenStore[key] = nil
    end)
  end
end

function MapFullSettings:tabMapGenSizeConfig()
  local shapeModeTexts = {
    [MapGenShape.Quad] = 'Square',
    [MapGenShape.Circle] = 'Circle',
    [MapGenShape.CircleCentered] = 'Centered Circle',
  }
  local shapeModePreview = shapeModeTexts[self.mapGenStore.shapeMode]
  ui.combo('Shape Optimization', shapeModePreview, function()
    table.forEach(shapeModeTexts, function(item, key)
      if ui.selectable(item, self.mapGenStore.shapeMode == key) then
        self.mapGenStore.shapeMode = key
      end
    end)
  end)
  if ui.itemHovered() then
    ui.tooltip(function()
      ui.text("Square:")
      ui.bullet()
      ui.text("Fills a square image")
      ui.bullet()
      ui.text("Optimal in combination with Full Map > Disable Rotation Margin")

      ui.text("Circle:")
      ui.bullet()
      ui.text("Fills a circle, allowing rotation without downscaling")
      ui.bullet()
      ui.text("Map might seem out of center")

      ui.text("Centered Circle:")
      ui.bullet()
      ui.text("Fills a circle while keeping the center of the map the same as in quadratic mode")
    end)
  end

  self.mapGenStore.imageSize = TooltipSlider('Image Size', self.mapGenStore.imageSize, 128, 8096, '%.0f px')
end

function MapFullSettings:tabMapGenBatchGeneration()
  SpacingSeparator()
  ui.text('Batch Generation')

  if ui.button('Create Batch Script File') then
    local fileName = self.mapGenerator:createBatchScript(self.mapGenStore)
    ui.toast(ui.Icons.Download, 'Batch script file created: ' .. fileName)
  end
  if ui.itemHovered() then
    ui.tooltip(function()
      ui.text("Creates a .bat file in the map_display app directory: apps/lua/map_display")
      ui.text(
        "This script contains the current parameters and can be used to generate (all) map images outside of the game.")
      ui.text("Each map widget creates its separate script file.")
    end)
  end

  if io.fileExists(self.mapGenerator:getBatchScriptPath()) then
    ui.sameLine()
    if ui.button('View in Folder') then
      os.showInExplorer(self.mapGenerator:getBatchScriptPath())
    end
  end
end

function MapFullSettings:tabFreeroamTeleport()
  SpacingSeparator()

  ui.text("Teleport")
  SpacingY()

  if ui.checkbox('Teleports enabled', self.freeroamStore.enableTeleports) then
    self.freeroamStore.enableTeleports = not self.freeroamStore.enableTeleports
    self:addChangedAttribute('enableTeleports')
  end

  -- teleport point colors
  if ui.colorButton(
        'Teleport Color',
        self.freeroamStore.teleportCircleColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.freeroamStore.teleportCircleColor = self.freeroamStore.teleportCircleColor
  end
  ui.sameLine()
  if ui.colorButton(
        'Teleport Outline',
        self.freeroamStore.teleportCircleOutline,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.freeroamStore.teleportCircleOutline = self.freeroamStore.teleportCircleOutline
  end
  ui.sameLine()
  ui.text('Teleport indicator colors')

  self.freeroamStore.teleportCircleRadius = TooltipSlider('Indicator Size', self.freeroamStore.teleportCircleRadius, 1,
    20, '%.0f px')
end

function MapFullSettings:tabDebug()
  local drawDebugLines = refbool(self.store.drawDebugLines)
  if ui.checkbox('Draw Debug Lines', drawDebugLines) then
    self.store.drawDebugLines = drawDebugLines.value
  end
  if ui.itemHovered() then
    ui.tooltip(function()
      ui.text("Red: map image boundaries")
      ui.text("Blue: image rotation area (area that could be used when different rotations are applied)")
      ui.text("Yellow: app edges")
    end)
  end
end

function MapFullSettings:initMapImage()
  local generatedMapPng = self.mapGenerator:getTrackMapPath() .. '.png'
  local generatedMapIni = self.mapGenerator:getTrackMapPath() .. '.ini'

  local noMapSelected =
      self.trackStore.fileImage == ''
      or self.trackStore.fileIni == ''
      or not io.fileExists(self.trackStore.fileImage)
      or not io.fileExists(self.trackStore.fileIni)

  local selectedMapIsGenerated =
      self.trackStore.fileImage == generatedMapPng
      and self.trackStore.fileIni == generatedMapIni

  -- try generate, reuses existing if hash matches
  if self.mapGenStore.enableMapGeneration then
    self:generateMap((noMapSelected and self.mapGenStore.selectGeneratedMap) or selectedMapIsGenerated)
  end

  if noMapSelected then
    if not self.mapGenStore.enableMapGeneration and self.mapGenStore.selectGeneratedMap then -- e.g. for batch generation files
      if io.exists(generatedMapPng) and io.exists(generatedMapIni) then
        self:setFile(generatedMapPng, generatedMapIni)
      else -- fallback if file does not exist
        self:setFile()
      end
    end

    -- explicitly select default map
    if not self.mapGenStore.selectGeneratedMap then
      self:setFile()
    end
  else
    -- use configured maps
    self:setFile(self.trackStore.fileImage, self.trackStore.fileIni)
  end
end

function MapFullSettings:generateMap(selectResult, forceGeneration)
  self.mapGenerator:generatePNG(
    self.mapGenStore,
    forceGeneration or false,
    selectResult
    and Bind(self.setFile, self)
    or function()
      -- do nothing
    end)
end

function MapFullSettings:setFile(fileImage, fileIni)
  -- move somewhere else
  if not fileImage or not fileIni then
    fileImage = ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\') .. '\\map.png'
    fileIni = ac.getFolder(ac.FolderID.ContentTracks) .. '\\' .. ac.getTrackFullID('\\') .. '\\data\\map.ini'
  end

  ac.debug(self.name .. ' fileImage', fileImage)
  ac.debug(self.name .. ' fileIni', fileIni)

  -- load and decode image manually, to be able to reload the same file after generation
  local imageRef = fileImage
  if io.exists(imageRef) then
    local imageData = io.load(imageRef)
    if imageData then
      imageRef = ui.decodeImage(imageData) or imageRef
    end
  end

  self.trackStore.fileImage = fileImage
  self.trackStore.fileIni = fileIni

  local mapData = ac.INIConfig.load(fileIni)
  local circleFit = mapData:get('PARAMETERS', 'CIRCLE_FIT', {})[1]

  self.mapInfo = {
    ref = imageRef,
    scaleFactor = mapData:get('PARAMETERS', 'SCALE_FACTOR', 1),
    offset = vec2(
      mapData:get('PARAMETERS', 'X_OFFSET', 1),
      mapData:get('PARAMETERS', 'Z_OFFSET', 1)),
    circleFit = (circleFit ~= nil and circleFit ~= '') and circleFit or nil
  }
end
