MapLegend = class('MapLegend')
function MapLegend:initialize(name, mapSettings) -- constructor
  self.name = name
  self.mapSettings = mapSettings

  -- widget specific persistent storage
  self.store = ac.storage({
    drawBackground = true,
    background = rgbm(0, 0, 0, 0.2),
    fontSize = 12,
    drawNumbers = true,
    removeLeadingHashs = true,
  }, name .. '_Legend')

  self.visible = false
  self.colors = {}
  self.timer = 0
end

function MapLegend:draw()
  if not self.visible then
    self.visible = true
    self.timer = math.huge
  end

  if table.nkeys(self.colors) == 0 then
    ui.text('No legend values.')
    return
  end

  local padding = 10
  local fontSize = self.store.fontSize

  local spacing = 3
  local textHeight = ui.measureDWriteText('A', fontSize, -1).y

  local size = vec2(
        ui.windowWidth(),
        (textHeight + spacing) * table.nkeys(self.colors) - spacing)
      :add(padding * 2)

  if self.store.drawBackground then
    display.rect({ pos = ui.getCursor():sub(padding), size = size, color = self.store.background })
  end

  table.forEach(self.colors, function(item, idx)
    local entryName, color, textColor = item.name, item.color, item.textColor

    -- circle
    local circleRadius = textHeight / 2
    local circlePosition = ui.getCursor():add(vec2(circleRadius, circleRadius + 1))
    ui.drawCircleFilled(circlePosition, circleRadius, color)
    ui.drawCircle(circlePosition, circleRadius, rgbm(0, 0, 0, 1), 36, 1)

    -- preview text
    if self.store.drawNumbers then
      ui.pushDWriteFont('consolas;Weight=Bold')
      local offset = ui.measureDWriteText(idx, fontSize, -1):scale(0.5)
      ui.dwriteDrawText(idx, fontSize, circlePosition:sub(offset), textColor)
      ui.popDWriteFont()
    end
    ui.beginOutline()

    -- entry name
    ui.pushDWriteFont('Segoe UI;Weight=Medium')
    ui.offsetCursorX(circleRadius * 2 + 5)
    ui.dwriteText(entryName, fontSize)
    ui.popDWriteFont()

    ui.endOutline(rgbm(0, 0, 0, ac.windowFading()))
  end)
end

function MapLegend:hide()
  self.visible = false
end

function MapLegend:drawSettings()
  ui.text('Window:')

  local fontSize = refnumber(self.store.fontSize)
  if ui.slider('Font Size', fontSize, 0, 64, '%.0f') then
    self.store.fontSize = fontSize.value
  end

  local drawNumbers = refbool(self.store.drawNumbers)
  if ui.checkbox('Draw Numbers', drawNumbers) then
    self.store.drawNumbers = drawNumbers.value
  end

  local removeLeadingHashs = refbool(self.store.removeLeadingHashs)
  if ui.checkbox('Remove Leading Hashs', removeLeadingHashs) then
    self.store.removeLeadingHashs = removeLeadingHashs.value
    self.timer = math.huge
  end

  local drawBackground = refbool(self.store.drawBackground)
  if ui.checkbox('Draw Background', drawBackground) then
    self.store.drawBackground = drawBackground.value
  end

  if drawBackground.value then
    local background = self.store.background
    if ui.colorPicker('Background', background, ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.PickerHueBar) then
      self.store.background = background
    end
  end
end

function MapLegend:update(dt)
  if not self.visible then
    return
  end

  self.timer = self.timer + dt

  if self.timer > 5 then
    self.colors = {}

    table.insert(self.colors, {
      name = 'Focused',
      color = self.mapSettings.store.specColor,
      textColor = self.mapSettings.store.specTextColor
    })

    if self.mapSettings.store.circleColorMode == CircleColorMode.Static then
      table.insert(self.colors, self:calculateStaticEntries())
    elseif self.mapSettings.store.circleColorMode == CircleColorMode.Position then
      table.forEach(self:calculatePositionEntries(), function(item)
        table.insert(self.colors, item)
      end)
    elseif self.mapSettings.store.circleColorMode == CircleColorMode.Class then
      table.forEach(self:calculateClassEntries(), function(item)
        table.insert(self.colors, item)
      end)
    end

    if self.mapSettings.store.stoppedCarsMode == StoppedCarsMode.Colored then
      table.insert(self.colors, {
        name = 'Stopped',
        color = self.mapSettings.store.stoppedCarsColor,
        textColor = self.mapSettings.store.stoppedCarsTextColor
      })
    end

    self.timer = 0
  end
end

---@return table
function MapLegend:calculateStaticEntries()
  return {
    name = 'Opponents',
    color = self.mapSettings.store.circleColorStatic,
    textColor = self.mapSettings.store.circleColorStaticText
  }
end

---@return table
function MapLegend:calculatePositionEntries()
  local colors = {
    {
      name = 'Ahead',
      color = self.mapSettings.store.positionAheadColor,
      textColor = self.mapSettings.store.positionAheadTextColor
    },
    {
      name = 'Behind',
      color = self.mapSettings.store.positionBehindColor,
      textColor = self.mapSettings.store.positionBehindTextColor
    },
    {
      name = 'Ahead Lapped',
      color = self.mapSettings.store.positionAheadLappedColor,
      textColor = self.mapSettings.store.positionAheadLappedTextColor
    },
    {
      name = 'Behind Lapped',
      color = self.mapSettings.store.positionBehindLappedColor,
      textColor = self.mapSettings.store.positionBehindLappedTextColor
    }
  }

  if self.mapSettings.store.positionHighlightLeader then
    table.insert(colors, 1, {
      name = 'Leader',
      color = self.mapSettings.store.positionHighlightColor,
      textColor = self.mapSettings.store.positionHighlightTextColor
    })
  end

  return colors
end

---@return table
function MapLegend:calculateClassEntries()
  local allCarTags = table.reduce(CARS, {}, function(data, item)
    table.forEach(item.carTags, function(tag)
      if not table.contains(data, tag:lower()) then
        table.insert(data, tag:lower())
      end
    end)
    return data
  end)

  local usedClasses = table.filter(self.mapSettings.classSettings.classes, function(item)
    return table.contains(allCarTags, item.class:lower())
  end)

  return table.map(usedClasses, function(item)
    return {
      name = self.store.removeLeadingHashs and item.class:gsub('^#', '') or item.class,
      color = item.color,
      textColor = item.textColor
    }
  end)
end
