require('widgets/map_progress_settings')

local zeroVector = vec2()

MapProgress = class('MapProgress')
function MapProgress:initialize(name) -- constructor
  self.name = name
  self.visible = false
  self.updateTimer = 0

  self.renderCars = table.clone(CARS, false)
  self.carData = {}
  self.carsByClass = nil
  self.carsByClassColor = nil
  self.hasTraffic = false

  self:initializeSettings()

  self:update(math.huge)
end

function MapProgress:initializeSettings()
  self.settings = MapProgressSettings(self.name)
end

function MapProgress:draw()
  self.visible = true

  local drawOffset = ui.getCursor()
  local contentSize = ui.windowSize():sub(drawOffset)

  display.rect({ pos = drawOffset, size = contentSize, color = self.settings.progress.backgroundColor })

  local sideRectMargin = self.settings.progress.sideIndicatorMargin

  local sideRectSize = vec2(self.settings.progress.sideIndicatorSize, contentSize.y - sideRectMargin * 2)
  local innerSize = vec2(contentSize.x - 2 * (sideRectSize.x + sideRectMargin), contentSize.y)
  local innerTopLeft = vec2(drawOffset.x + (sideRectSize.x + sideRectMargin), drawOffset.y)
  local innerBotRight = vec2(innerTopLeft.x + innerSize.x, innerTopLeft.y + (sideRectSize.y + 2 * sideRectMargin))

  -- sector positions
  local splits = ac.getSim().lapSplits
  for i = 0, #splits - 1 do
    -- do not draw start finish
    if splits[i] > 0.1 then
      display.rect({
        pos = vec2(splits[i] * innerSize.x - self.settings.progress.sectorWidth / 2 + drawOffset.x, drawOffset.y),
        size = vec2(self.settings.progress.sectorWidth, contentSize.y),
        color = self.settings.progress.sectorColor
      })
    end
  end

  -- arrows
  ui.pushClipRect(innerTopLeft, innerBotRight)
  self:drawArrows({
    drawOffset = innerTopLeft,
    contentSize = innerSize
  })
  ui.popClipRect()

  -- left/right indicators
  local indicatorColor = self:calculateMapColor()
  display.rect({ pos = drawOffset:clone():add(sideRectMargin), size = sideRectSize, color = indicatorColor })
  display.rect({
    pos = vec2(innerBotRight.x, drawOffset.y + sideRectMargin),
    size = sideRectSize,
    color = indicatorColor
  })
end

function MapProgress:drawArrows(posCalculationParameters)
  ui.pushDWriteFont('Consolas;Weight=Bold')

  if self.renderCars ~= nil then
    -- draw special traffic nodes
    if self.hasTraffic and self.settings.freeroamStore.trafficMode == TrafficMode.Custom then
      local trafficColors = {
        color = self.settings.freeroamStore.trafficCircleColor,
        textColor = rgbm.colors.transparent,
      }
      for i = 0, #self.renderCars - 1 do
        local car = self.renderCars[#self.renderCars - i]
        if car.id ~= FOCUSED_CAR.id
            and self:validateCar(car)
            and self.carData[car.id].isTraffic then
          self:drawArrow(
            car,
            posCalculationParameters,
            self:calculateCircleRadius(car, self.settings.freeroamStore.trafficCircleRadius),
            trafficColors)
        end
      end
    end
    -- draw drivers
    for i = 0, #self.renderCars - 1 do
      local car = self.renderCars[#self.renderCars - i]
      if car.id ~= FOCUSED_CAR.id
          and self:validateCar(car)
          and not self.carData[car.id].isTraffic then
        self:drawArrow(
          car,
          posCalculationParameters,
          self:calculateCircleRadius(car, self.settings.store.circleRadius))
      end
    end
  end
  -- draw focused car
  if self:validateCar(FOCUSED_CAR) then
    self:drawArrow(
      FOCUSED_CAR,
      posCalculationParameters,
      self:calculateCircleRadius(FOCUSED_CAR, self.settings.store.specCircleRadius),
      { color = self.settings.store.specColor, textColor = self.settings.store.specTextColor })
  end

  ui.popDWriteFont()
end

function MapProgress:validateCar(car)
  return car
      and car.isActive
      and (self.settings.store.pitLaneMode ~= ProgressPitLaneMode.Hide or not car.inPitLane or car == FOCUSED_CAR)
      and self.carData[car.id]
      and not self.carData[car.id].hide
end

function MapProgress:drawArrow(car, posCalculationParameters, circleRadius, colors)
  local carData = self.carData[car.id]

  -- optional parameter fill in
  if not colors and carData.circleColors then
    colors = carData.circleColors
  elseif not colors then
    return
  end

  -- position calculation
  local pos2 = self:calculateCirclePositionOnMap(car, posCalculationParameters)
  local dir2 = self:calculateCircleScaledDirectionOnMap(car, posCalculationParameters, circleRadius)
  local dir2Orth = vec2(dir2.y, -dir2.x)

  -- render position
  ui.drawImageQuad(self.settings.store.circleImage,
    (pos2 - dir2):sub(dir2Orth),
    (pos2 - dir2):add(dir2Orth),
    (pos2 + dir2):add(dir2Orth),
    (pos2 + dir2):sub(dir2Orth),
    colors.color)

  if carData.text then
    local fontOffset = self.settings.store.circleImage == CircleImage.Pin and dir2:scale(1 / 6) or zeroVector
    local fontCenterPos = (pos2 - fontOffset) -- dir2 is already scaled to circleRadius
        :sub(ui.measureDWriteText(carData.text, circleRadius):scale(0.5))
    ui.dwriteDrawText(carData.text, circleRadius, fontCenterPos, colors.textColor)
  end
end

function MapProgress:calculateCircleRadius(car, circleRadius)
  return (self.settings.store.pitLaneMode == ProgressPitLaneMode.Scale and car.inPitLane)
      and (self.settings.store.pitLaneOffsetScale / 100) * circleRadius
      or circleRadius
end

function MapProgress:calculateCirclePositionOnMap(car, posCalculationParameters)
  return vec2(car.lapProgress,
        self.settings.store.pitLaneMode == ProgressPitLaneMode.Scale and car.inPitLane and 0.8 or 0.5)
      :mul(posCalculationParameters.contentSize)
      :add(posCalculationParameters.drawOffset)
end

function MapProgress:calculateCircleScaledDirectionOnMap(car, posCalculationParameters, circleRadius)
  return vec2(circleRadius, 0)
end

function MapProgress:calculateMapColor()
  return self.settings.store.colorByFlag
      and MapFlagColors[ac.getSim().raceFlagType]
      or rgbm.colors.white
end

function MapProgress:hide()
  self.visible = false
end

function MapProgress:drawSettings()
  self.settings:draw()
end

function MapProgress:update(dt)
  if not self.visible then
    return
  end

  self.updateTimer = self.updateTimer + dt
  if self.updateTimer > 0.5 then
    self:updateCars()
    self:lazyUpdate()
    self.updateTimer = 0
  end
end

function MapProgress:updateCars()
  table.sort(self.renderCars, function(a, b)
    local comp = Car.activeStateComparison(a, b)
    if comp == 0 then
      return Car.getComparatorForSession()(a, b)
    end
    return comp < 0
  end)
end

function MapProgress:lazyUpdate()
  --init carData if not done yet
  if table.nkeys(self.carData) ~= table.nkeys(CARS) then
    table.forEach(CARS, function(car)
      self.carData[car.id] = {
        isTraffic = false,
        hide = false,
        classInfo = nil,
        classColors = { color = rgbm.colors.gray, textColor = rgbm.colors.white },
        positionColors = { color = rgbm.colors.gray, textColor = rgbm.colors.white },
        circleColors = { color = rgbm.colors.gray, textColor = rgbm.colors.white },
        text = car.racePosition,
      }
    end)
  end

  -- adds classInfo to all cars and keeps it up to date
  local classesUpdated = self:updateCarClasses()

  local colors = nil
  local colorKey = nil

  if self.settings.store.circleColorMode == CircleColorMode.Class then
    self:calculateClassColors(classesUpdated)
    colorKey = 'classColors'
  elseif self.settings.store.circleColorMode == CircleColorMode.Position then
    self:calculatePositionColors()
    colorKey = 'positionColors'
  else
    colors = {
      color = self.settings.store.circleColorStatic,
      textColor = self.settings.store.circleColorStaticText
    }
  end

  self:calculateTraffic()

  -- calculate visible colors for every car
  table.forEach(CARS, function(car)
    self.carData[car.id].circleColors = self:calculateDynamicColor(car, colors or self.carData[car.id][colorKey])
  end)

  self:updateSortedCarsByClass()
  self:updateSortedCarsByClassColor()

  self:updateCarDataTexts()
end

function MapProgress:calculateClassOfCar(car)
  local lowerCaseCarTags = table.map(car.carTags, function(item)
    return item:lower()
  end)

  return table.findFirst(self.settings:getClasses(), function(item)
    return table.contains(lowerCaseCarTags, item.class:lower())
  end) or NoClass
end

function MapProgress:updateCarClasses()
  if self.settings:attributeChanged('classes') then
    table.forEach(CARS, function(car)
      self.carData[car.id].classInfo = self:calculateClassOfCar(car)
    end)
    return true
  end
  return false
end

function MapProgress:updateSortedCarsByClass(classesUpdated)
  if not self.carsByClass or classesUpdated then
    self.carsByClass = {}
    table.forEach(CARS, function(car)
      local class = self.carData[car.id].classInfo.class

      if not self.carsByClass[class] then
        self.carsByClass[class] = {}
      end

      table.insert(self.carsByClass[class], car)
    end)
  end

  table.forEach(self.carsByClass, function(item)
    table.sort(item, function(a, b)
      return Car.compareRacePosition(a, b) < 0
    end)
  end)
end

function MapProgress:updateSortedCarsByClassColor(classesUpdated)
  if not self.carsByClassColor or classesUpdated then
    self.carsByClassColor = {}
    table.forEach(CARS, function(car)
      local classColor = stringify({ self.carData[car.id].classInfo.color, self.carData[car.id].classInfo.textColor })

      if not self.carsByClassColor[classColor] then
        self.carsByClassColor[classColor] = {}
      end

      table.insert(self.carsByClassColor[classColor], car)
    end)
  end

  table.forEach(self.carsByClassColor, function(item)
    table.sort(item, function(a, b)
      return Car.compareRacePosition(a, b) < 0
    end)
  end)
end

function MapProgress:calculatePositionColors()
  if self.settings.store.circleColorMode == CircleColorMode.Position then
    local sessionType = ac.getSim().raceSessionType
    table.forEach(CARS, function(car)
      self.carData[car.id].positionColors = self:calculatePositionColor(car, sessionType)
    end)
  end
end

function MapProgress:calculatePositionColor(car, sessionType)
  if car.racePosition == 1 and self.settings.store.positionHighlightLeader then
    return {
      color = self.settings.store.positionHighlightColor,
      textColor = self.settings.store.positionHighlightTextColor
    }
  elseif sessionType == ac.SessionType.Race then
    local lapDiff = car.totalLaps - FOCUSED_CAR.totalLaps
    if lapDiff < -1 then
      return {
        color = self.settings.store.positionBehindLappedColor,
        textColor = self.settings.store.positionBehindLappedTextColor
      }
    elseif lapDiff > 1 then
      return {
        color = self.settings.store.positionAheadLappedColor,
        textColor = self.settings.store.positionAheadLappedTextColor
      }
    elseif lapDiff < 0 then
      return {
        color = self.settings.store.positionBehindColor,
        textColor = self.settings.store.positionBehindTextColor
      }
    else
      return {
        color = self.settings.store.positionAheadColor,
        textColor = self.settings.store.positionAheadTextColor
      }
    end
  else
    return car.racePosition > FOCUSED_CAR.racePosition
        and {
          color = self.settings.store.positionBehindColor,
          textColor = self.settings.store.positionBehindTextColor
        }
        or {
          color = self.settings.store.positionAheadColor,
          textColor = self.settings.store.positionAheadTextColor
        }
  end
end

function MapProgress:calculateClassColors(classesUpdated)
  if self.settings.store.circleColorMode == CircleColorMode.Class and classesUpdated then
    ac.debug(self.name .. ' class update', ac.getSim().time)
    table.forEach(CARS, function(car)
      local class = self.carData[car.id].classInfo
      self.carData[car.id].classColors = { color = class.color, textColor = class.textColor }
    end)
  end
end

function MapProgress:calculateDynamicColor(car, colors)
  local circleColor, textColor = colors.color, colors.textColor

  if self.settings.driverSettings.drivers[car.driver.name] and self.settings.driverSettings.drivers[car.driver.name].hidden then
    return nil
  end

  if self.settings.store.stoppedCarsMode ~= StoppedCarsMode.Default
      and car.speed < self.settings.store.stoppedCarsThreshold then
    if self.settings.store.stoppedCarsMode == StoppedCarsMode.Colored then
      circleColor = self.settings.store.stoppedCarsColor
      textColor = self.settings.store.stoppedCarsTextColor
    elseif self.settings.store.stoppedCarsMode == StoppedCarsMode.Transparent then
      circleColor = rgbm(
        circleColor.r, circleColor.g, circleColor.b, circleColor.mult * self.settings.store.stoppedCarsAlpha)
      textColor = rgbm(
        textColor.r, textColor.g, textColor.b, textColor.mult * self.settings.store.stoppedCarsAlpha)
    elseif self.settings.store.stoppedCarsMode == StoppedCarsMode.Hide then
      return nil
    end
  end

  return { color = circleColor, textColor = textColor }
end

function MapProgress:updateCarDataTexts()
  table.forEach(CARS, function(car)
    local carData = self.carData[car.id]

    if self.settings.store.circleTextMode == CircleTextMode.None then
      carData.text = nil
    elseif self.settings.store.circleTextMode == CircleTextMode.Position then
      carData.text = car.racePosition > 0 and car.racePosition or ''
    elseif self.settings.store.circleTextMode == CircleTextMode.ClassPosition then
      local pos = table.indexOf(self.carsByClass[carData.classInfo.class], car)
      carData.text = pos > 0 and pos or ''
    elseif self.settings.store.circleTextMode == CircleTextMode.ClassColorPos then
      local classColor = stringify({ carData.classInfo.color, carData.classInfo.textColor })
      local pos = table.indexOf(self.carsByClassColor[classColor], car)
      carData.text = pos > 0 and pos or ''
    elseif self.settings.store.circleTextMode == CircleTextMode.Initials then
      carData.text = car.driver.initials
    elseif self.settings.store.circleTextMode == CircleTextMode.Id then
      carData.text = car.id
    end
  end)
end

function MapProgress:calculateTraffic()
  if self.settings:attributeChanged('traffic') then
    ac.debug(self.name .. ' traffic update', ac.getSim().time)

    self.hasTraffic = false

    table.forEach(CARS, function(car)
      if self.settings.freeroamStore.trafficMode == TrafficMode.Show then
        self.carData[car.id].isTraffic = false
        self.carData[car.id].hide = false
      else
        self.carData[car.id].isTraffic = self:isTrafficCar(car)

        -- set other properties only for traffic cars
        if self.carData[car.id].isTraffic then
          self.hasTraffic = true
          self.carData[car.id].hide = self.settings.freeroamStore.trafficMode == TrafficMode.Hide
        end
      end
    end)
  end
end

function MapProgress:isTrafficCar(car)
  return self.settings.freeroamStore.trafficAutoDetection and car.isHidingLabels
end
