require('helpers/map_generation')
require('widgets/settings/map_driver_settings')
require('widgets/settings/map_class_settings')

MapProgressSettings = class('MapProgressSettings')

MapProgressSettings.defaultGeneral = require('widgets/settings/general_defaults')
MapProgressSettings.defaultProgress = require('widgets/settings/progress_defaults')
MapProgressSettings.defaultMapGen = require('widgets/settings/map_gen_defaults')
MapProgressSettings.defaultTrack = require('widgets/settings/track_defaults')
MapProgressSettings.defaultFreeroam = require('widgets/settings/freeroam_defaults')

function MapProgressSettings.defaultGeneralProgress()
  local settings = MapProgressSettings.defaultGeneral()
  settings.circleImage = CircleImage.Circle
  return settings
end

function MapProgressSettings:initialize(name) -- constructor
  self.name = name

  self.store = ac.storage(self.defaultGeneralProgress(), name .. '_Settings')

  self.progress = ac.storage(self.defaultProgress(), name .. '_Progress')

  self.driverSettings = MapDriverSettings(name)

  self.classSettings = MapClassSettings(name)

  self.freeroamStore = ac.storage(self.defaultFreeroam(), name .. '_Freeroam')

  self.changedAttributes = { 'traffic' }
end

function MapProgressSettings:draw()
  ui.childWindow('main', vec2(450, 660), function()
    return self:mainContent()
  end)
end

function MapProgressSettings:attributeChanged(attr)
  local changed = table.contains(self.changedAttributes, attr)

  if changed then
    table.removeItem(self.changedAttributes, attr)
  end

  return changed or self.classSettings:attributeChanged(attr)
end

function MapProgressSettings:addChangedAttribute(attr)
  if not table.contains(self.changedAttributes, attr) then
    table.insert(self.changedAttributes, attr)
  end
end

function MapProgressSettings:mainContent()
  ui.tabBar('tabs', ui.TabBarFlags.FittingPolicyScroll + ui.TabBarFlags.TabListPopupButton, function()
    ui.tabItem('General', Bind(self.tabMain, self))
    ui.tabItem('Progress Map', Bind(self.tabProgress, self))
    ui.tabItem('Drivers', Bind(self.tabDrivers, self))
    ui.tabItem('Classes', Bind(self.tabClasses, self))
    ui.tabItem('Freeroam', Bind(self.tabFreeroam, self))
    return nil
  end)
end

function MapProgressSettings:tabMain()
  ui.text("General")
  SpacingY()

  local circleTextModeTexts = {
    [CircleTextMode.None] = 'No Text',
    [CircleTextMode.Position] = 'Position',
    [CircleTextMode.ClassPosition] = 'Position in Class',
    [CircleTextMode.ClassColorPos] = 'Position by Class Color',
    [CircleTextMode.Initials] = 'Initials',
    [CircleTextMode.Id] = 'Car ID',
  }
  local circleTextPreview = circleTextModeTexts[self.store.circleTextMode]
  ui.combo('Text Mode', circleTextPreview, function()
    table.forEach(circleTextModeTexts, function(item, key)
      if ui.selectable(item, self.store.circleTextMode == key) then
        self.store.circleTextMode = key
      end
    end)
  end)
  -- circle image
  local circleImageTexts = {
    [CircleImage.Pin] = 'Pin',
    [CircleImage.Circle] = 'Circle',
    [CircleImage.CircleFlat] = 'Flat Circle',
  }
  local circleImagePreview = circleImageTexts[self.store.circleImage]
  ui.combo('Marker Image', circleImagePreview, function()
    table.forEach(circleImageTexts, function(item, key)
      if ui.selectable(item, self.store.circleImage == key) then
        self.store.circleImage = key
      end
    end)
  end)
  -- circle radius
  self.store.circleRadius = TooltipSlider('Marker Size', self.store.circleRadius, 1, 64, '%.0f px')

  if ui.checkbox('Color by flag', self.store.colorByFlag) then
    self.store.colorByFlag = not self.store.colorByFlag
  end

  SpacingSeparator()

  -- spectated driver overrides
  ui.text('Spectated Car')
  SpacingY()
  if ui.colorButton(
        'Selected Car Color',
        self.store.specColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.store.specColor = self.store.specColor
  end
  ui.sameLine()
  if ui.colorButton(
        'Selected Car Text Color',
        self.store.specTextColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.store.specTextColor = self.store.specTextColor
  end
  ui.sameLine()
  ui.text('Spectated Car Color')

  self.store.specCircleRadius = TooltipSlider('Spec Marker Size', self.store.specCircleRadius, 1, 64, '%.0f px')

  SpacingSeparator()

  self:tabMainStoppedCarsMode()

  SpacingSeparator()

  self:tabMainPitLaneMode()

  SpacingSeparator()

  self:tabMainColorMode()

  SpacingSeparator()

  -- Reset
  if ui.button('Reset to default') then
    table.forEach(self.defaultGeneralProgress(), function(item, key)
      self.store[key] = nil
    end)
  end
end

function MapProgressSettings:tabMainStoppedCarsMode()
  -- stopped cars settings
  ui.text('Stopped Cars')
  SpacingY()
  local stoppedCarsModeTexts = {
    [StoppedCarsMode.Default] = 'Show',
    [StoppedCarsMode.Hide] = 'Hide',
    [StoppedCarsMode.Transparent] = 'Transparent',
    [StoppedCarsMode.Colored] = 'Custom Color',
  }
  local stoppedCarsPreview = stoppedCarsModeTexts[self.store.stoppedCarsMode]
  ui.combo('Stopped Cars Mode', stoppedCarsPreview, function()
    table.forEach(stoppedCarsModeTexts, function(item, key)
      if ui.selectable(item, self.store.stoppedCarsMode == key) then
        self.store.stoppedCarsMode = key
      end
    end)
  end)
  -- dynamic mode content
  if self.store.stoppedCarsMode == StoppedCarsMode.Colored then
    if ui.colorButton(
          'Stopped Car Color',
          self.store.stoppedCarsColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.stoppedCarsColor = self.store.stoppedCarsColor
    end
    ui.sameLine()
    if ui.colorButton(
          'Stopped Car Text Color',
          self.store.stoppedCarsTextColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.stoppedCarsTextColor = self.store.stoppedCarsTextColor
    end
    ui.sameLine()
    ui.text('Stopped Car Color')
  elseif self.store.stoppedCarsMode == StoppedCarsMode.Transparent then
    self.store.stoppedCarsAlpha = TooltipSlider('Stopped Cars Alpha', self.store.stoppedCarsAlpha, 0.0, 1.0, '%.1f px')
  end

  self.store.stoppedCarsThreshold = TooltipSlider('Stopped Threshold', self.store.stoppedCarsThreshold, 0.0, 20.0,
    '%.1f m/s')
end

function MapProgressSettings:tabMainPitLaneMode()
  ui.text('Pit Lane')
  SpacingY()

  local pitLaneModeTexts = {
    [ProgressPitLaneMode.Show] = 'Show',
    [ProgressPitLaneMode.Hide] = 'Hide',
    [ProgressPitLaneMode.Scale] = 'Scale',
  }
  local pitLaneModePreview = pitLaneModeTexts[self.store.pitLaneMode]
  ui.combo('Pit Lane Mode', pitLaneModePreview, function()
    table.forEach(pitLaneModeTexts, function(item, key)
      if ui.selectable(item, self.store.pitLaneMode == key) then
        self.store.pitLaneMode = key
      end
    end)
  end)

  if self.store.pitLaneMode == ProgressPitLaneMode.Scale then
    self.store.pitLaneOffsetScale = TooltipSlider(
      'Marker Scale', self.store.pitLaneOffsetScale, 0, 200, '%.0f%%')
  end
end

function MapProgressSettings:tabMainColorMode()
  -- color modes
  ui.text('Color Mode')
  SpacingY()
  local colorModeTexts = {
    [CircleColorMode.Static] = 'Static',
    [CircleColorMode.Position] = 'Position Based',
    [CircleColorMode.Class] = 'Class Based',
  }
  local colorModePreview = colorModeTexts[self.store.circleColorMode]
  ui.combo('Color Mode', colorModePreview, function()
    table.forEach(colorModeTexts, function(item, key)
      if ui.selectable(item, self.store.circleColorMode == key) then
        self.store.circleColorMode = key
        if key == CircleColorMode.Class then
          self:addChangedAttribute('classes')
        end
      end
    end)
  end)
  -- dynamic mode content
  if self.store.circleColorMode == CircleColorMode.Static then
    if ui.colorButton(
          'Car Color',
          self.store.circleColorStatic,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.circleColorStatic = self.store.circleColorStatic
    end
    ui.sameLine()
    if ui.colorButton(
          'Car Text Color',
          self.store.circleColorStaticText,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.circleColorStaticText = self.store.circleColorStaticText
    end
    ui.sameLine()
    ui.text('Car Color')
  elseif self.store.circleColorMode == CircleColorMode.Position then
    ui.columns(2, false)

    if ui.checkbox('Highlight leader', self.store.positionHighlightLeader) then
      self.store.positionHighlightLeader = not self.store.positionHighlightLeader
    end

    ui.nextColumn()

    -- highlight
    if ui.colorButton(
          'Highlight Color',
          self.store.positionHighlightColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionHighlightColor = self.store.positionHighlightColor
    end
    ui.sameLine()
    if ui.colorButton(
          'Highlight Text Color',
          self.store.positionHighlightTextColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionHighlightTextColor = self.store.positionHighlightTextColor
    end
    ui.sameLine()
    ui.text('Highlight')

    ui.nextColumn()

    -- ahead
    if ui.colorButton(
          'Ahead Color',
          self.store.positionAheadColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionAheadColor = self.store.positionAheadColor
    end
    ui.sameLine()
    if ui.colorButton(
          'Ahead Text Color',
          self.store.positionAheadTextColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionAheadTextColor = self.store.positionAheadTextColor
    end
    ui.sameLine()
    ui.text('Ahead')

    ui.nextColumn()

    -- behind
    if ui.colorButton(
          'Behind Color',
          self.store.positionBehindColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionBehindColor = self.store.positionBehindColor
    end
    ui.sameLine()
    if ui.colorButton(
          'Behind Text Color',
          self.store.positionBehindTextColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionBehindTextColor = self.store.positionBehindTextColor
    end
    ui.sameLine()
    ui.text('Behind')

    ui.nextColumn()

    -- ahead lapped
    if ui.colorButton(
          'Ahead Lapped Color',
          self.store.positionAheadLappedColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionAheadLappedColor = self.store.positionAheadLappedColor
    end
    ui.sameLine()
    if ui.colorButton(
          'Ahead Lapped Text Color',
          self.store.positionAheadLappedTextColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionAheadLappedTextColor = self.store.positionAheadLappedTextColor
    end
    ui.sameLine()
    ui.text('Ahead Lapped')

    ui.nextColumn()

    -- behind lapped
    if ui.colorButton(
          'Behind Lapped Color',
          self.store.positionBehindLappedColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionBehindLappedColor = self.store.positionBehindLappedColor
    end
    ui.sameLine()
    if ui.colorButton(
          'Behind Lapped Text Color',
          self.store.positionBehindLappedTextColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.store.positionBehindLappedTextColor = self.store.positionBehindLappedTextColor
    end
    ui.sameLine()
    ui.text('Behind Lapped')

    ui.columns()
  elseif self.store.circleColorMode == CircleColorMode.Class then
    ui.textColored("Classes are configured in the 'Classes' tab", rgbm.colors.yellow)
  end

  SpacingY()

  CopyPaste('colorMode', 'Color mode config', self.store,
    { 'circleColorMode', 'circleColorStatic', 'circleColorStaticText',
      'positionHighlightLeader', 'positionHighlightColor', 'positionHighlightTextColor',
      'positionBehindColor', 'positionBehindTextColor',
      'positionBehindLappedColor', 'positionBehindLappedTextColor',
      'positionAheadColor', 'positionAheadTextColor',
      'positionAheadLappedColor', 'positionAheadLappedTextColor' })
end

function MapProgressSettings:tabProgress()
  ui.text("Progress Map")
  SpacingY()

  if ui.colorButton(
        'Background Color',
        self.progress.backgroundColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.progress.backgroundColor = self.progress.backgroundColor
  end
  ui.sameLine()
  ui.text('Background Color')

  SpacingSeparator()

  self.progress.sideIndicatorSize = TooltipSlider(
    'Side Indicator Width', self.progress.sideIndicatorSize, 0, 128, '%.0f px')
  self.progress.sideIndicatorMargin = TooltipSlider(
    'Side Indicator Margin', self.progress.sideIndicatorMargin, 0, 24, '%.0f px')

  SpacingSeparator()

  if ui.colorButton(
        'Sector Color',
        self.progress.sectorColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
    self.progress.sectorColor = self.progress.sectorColor
  end
  ui.sameLine()
  ui.text('Sector Color')

  self.progress.sectorWidth = TooltipSlider(
    'Sector Width', self.progress.sectorWidth, 0, 128, '%.0f px')

  SpacingSeparator()

  -- Reset
  if ui.button('Reset to default') then
    table.forEach(self.defaultProgress(), function(item, key)
      self.progress[key] = nil
    end)
  end
end

function MapProgressSettings:tabClasses()
  if self.store.circleColorMode ~= CircleColorMode.Class then
    SpacingY()

    ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.orange)
    ui.textWrapped(
      "Color mode 'Class Based' is currently not selected for this widget. This configuration is only relevant for that mode.",
      ui.availableSpaceX())
    ui.popStyleColor()

    SpacingY()

    ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.gray)
  end

  self.classSettings:tabClasses()

  if self.store.circleColorMode ~= CircleColorMode.Class then
    ui.popStyleColor()
  end
end

function MapProgressSettings:getClasses()
  return self.classSettings:getClasses()
end

function MapProgressSettings:tabDrivers()
  self.driverSettings:tabDrivers()
end

function MapProgressSettings:tabFreeroam()
  ui.text("Traffic")
  SpacingY()
  local trafficModeTexts = {
    [TrafficMode.Show] = 'Show',
    [TrafficMode.Hide] = 'Hide',
    [TrafficMode.Custom] = 'Custom Display',
  }
  local trafficModePreview = trafficModeTexts[self.freeroamStore.trafficMode]
  ui.combo('Traffic Mode', trafficModePreview, function()
    table.forEach(trafficModeTexts, function(item, key)
      if ui.selectable(item, self.freeroamStore.trafficMode == key) then
        self.freeroamStore.trafficMode = key
        self:addChangedAttribute('traffic')
      end
    end)
  end)
  if self.freeroamStore.trafficMode == TrafficMode.Custom then
    self.freeroamStore.trafficCircleRadius = TooltipSlider('Traffic Size', self.freeroamStore.trafficCircleRadius, 1,
      50, '%.0f px')
    if ui.colorButton(
          'Traffic Color',
          self.freeroamStore.trafficCircleColor,
          ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar) then
      self.freeroamStore.trafficCircleColor = self.freeroamStore.trafficCircleColor
    end
    ui.sameLine()
    ui.text('Traffic Color')
  end

  self:tabFreeroamTeleport()

  SpacingSeparator()

  -- Reset
  if ui.button('Reset to default') then
    table.forEach(self.defaultFreeroam(), function(item, key)
      self.freeroamStore[key] = nil
    end)
  end
end

function MapProgressSettings:tabFreeroamTeleport()
  -- hide
end
