local defaultClasses = require('widgets/settings/class_defaults')

MapClassSettings = class('MapClassSettings')
function MapClassSettings:initialize(name) -- constructor
  self.classesStore = ac.storage({
    -- class configs
    classes = stringify(defaultClasses()),
  }, name .. '_Classes')

  self.changedAttributes = { 'classes' }
  self.classes = self:parseClasses()
end

function MapClassSettings:attributeChanged(attr)
  local changed = table.contains(self.changedAttributes, attr)
  table.removeItem(self.changedAttributes, attr)
  return changed
end

function MapClassSettings:addChangedAttribute(attr)
  if not table.contains(self.changedAttributes, attr) then
    table.insert(self.changedAttributes, attr)
  end
end

function MapClassSettings:tabClasses()
  -- Active classes
  ui.text('Active Classes')
  SpacingY()

  ac.debug('Classes', self.classesStore.classes)

  ui.beginGroup()
  ui.columns(4, false)
  ui.setColumnWidth(0, 50)
  ui.setColumnWidth(1, ui.modernButton and 300 or 250)
  ui.setColumnWidth(2, (ui.textLineHeightWithSpacing() + 15) * 2)

  local tempClasses = self:parseClasses()
  if type(tempClasses) ~= "table" then
    return
  end

  local classCount = table.nkeys(tempClasses)
  table.forEach(tempClasses, function(item, key)
    if item ~= nil then
      -- up/down
      if ui.arrowButton(
            'Up##' .. key,
            ui.Direction.Up,
            vec2(15, ui.frameHeight()),
            key > 1 and ui.ButtonFlags.None or ui.ButtonFlags.Disabled) then
        table.remove(tempClasses, key)
        table.insert(tempClasses, key - 1, item)
      end
      ui.sameLine()
      if ui.arrowButton(
            'Down##' .. key,
            ui.Direction.Down,
            vec2(15, ui.frameHeight()),
            key < classCount and ui.ButtonFlags.None or ui.ButtonFlags.Disabled) then
        table.remove(tempClasses, key)
        table.insert(tempClasses, key + 1, item)
      end
      ui.nextColumn()

      -- text
      ui.setNextItemWidth(ui.availableSpaceX())
      local classText, _, enterPressed = ui.inputText('##name' .. key, item.class)
      if ui.itemHovered() then
        ui.tooltip(function()
          ui.text("Case insensitive")
          ui.text("Press enter to confirm changes")
        end)
      end
      if enterPressed then
        item.class = classText
      end
      ui.nextColumn()

      -- color preview
      ui.colorButton(
        'Color##' .. key,
        item.color,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar)
      ui.sameLine()
      ui.colorButton(
        'Text Color##' .. key,
        item.textColor,
        ui.ColorPickerFlags.AlphaBar + ui.ColorPickerFlags.AlphaPreview + ui.ColorPickerFlags.PickerHueBar)

      ui.nextColumn()

      -- buttons
      if RemoveButton(key) then
        table.remove(tempClasses, key)
      end
      ui.nextColumn()
    end
  end)

  ui.endGroup()
  ui.columns()

  SpacingSeparator()
  ui.text('SubStanding (Extended) export')
  if ui.button('Export Class Tags') then
    ui.setClipboardText(table.concat(table.map(tempClasses, function(item)
      return item.class
    end), ','))
    ui.toast(ui.Icons.Clipboard, 'Class tags copied!')
  end
  ui.sameLine()
  if ui.button('Export Class Colors') then
    ui.setClipboardText(table.concat(table.map(tempClasses, function(item)
      return RgbmToHex(item.color)
    end), ','))
    ui.toast(ui.Icons.Clipboard, 'Class colors copied!')
  end
  -- Add Classes
  SpacingSeparator()
  ui.setNextItemWidth(ui.availableSpaceX())
  ui.combo('##newclass', 'Add Class from current vehicles', function()
    local existingClasses = table.map(tempClasses, function(item)
      return item.class:lower()
    end)

    local tags = table.reduce(CARS, {}, function(data, item)
      table.forEach(item.carTags, function(tag)
        if not table.contains(existingClasses, tag:lower()) then
          data[tag] = (data[tag] or 0) + 1
        end
      end)
      return data
    end)

    local tagCountsList = table.map(tags, function(item, key)
      return { tag = key, count = item }
    end)

    table.sort(tagCountsList, function(a, b)
      return a.count > b.count
    end)

    -- draw calculated tags
    table.forEach(tagCountsList, function(item)
      if ui.selectable(string.format('%s (%d vehicles)', item.tag, item.count)) then
        table.insert(tempClasses,
          { class = item.tag, color = rgbm.colors.gray:clone(), textColor = rgbm.colors.white:clone() })
      end
    end)
  end)

  self:setClasses(tempClasses)

  -- Reset
  if ui.button('Reset to default') then
    self:setClasses(defaultClasses())
  end
  ui.sameLine()

  if CopyPaste('classes', 'Class definitions', self.classesStore, { 'classes' }, function()
        self.classes = self:parseClasses()
        self:addChangedAttribute('classes')
      end) then
    return
  end
end

function MapClassSettings:parseClasses()
  -- in csp 0.1.79 tryParse returns nil in case of error
  return stringify.tryParse(self.classesStore.classes, {}, defaultClasses()) or defaultClasses()
end

function MapClassSettings:getClasses()
  return self.classes
end

function MapClassSettings:setClasses(classes)
  if not TableCompare(classes, stringify.tryParse(self.classesStore.classes, {})) then
    self.classes = classes
    self.classesStore.classes = stringify(classes)

    self:addChangedAttribute('classes')
  end
end
