local defaultDriver = require('widgets/settings/driver_defaults')

MapDriverSettings = class('MapDriverSettings')
function MapDriverSettings:initialize(name) -- constructor
  self.driversStore = ac.storage({
    -- driver configs
    drivers = stringify({}),
  }, name .. '_Drivers')

  self.changedAttributes = { 'drivers' }
  self.drivers = self:parseDrivers()
end

function MapDriverSettings:attributeChanged(attr)
  local changed = table.contains(self.changedAttributes, attr)
  table.removeItem(self.changedAttributes, attr)
  return changed
end

function MapDriverSettings:addChangedAttribute(attr)
  if not table.contains(self.changedAttributes, attr) then
    table.insert(self.changedAttributes, attr)
  end
end

function MapDriverSettings:tabDrivers()
  -- Active drivers
  ui.text('Driver Settings')
  SpacingY()

  ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.yellow)
  ui.textWrapped(
    "Hidden drivers will not be excluded from the position count, as it's not expected that they actively participate. So they should be last anyways.",
    ui.availableSpaceX())
  ui.popStyleColor()

  SpacingSeparator()

  ac.debug('Drivers', self.driversStore.drivers)

  local tempDrivers = self:parseDrivers()
  if type(tempDrivers) ~= "table" then
    return
  end

  local existingDrivers = table.map(tempDrivers, function(item, key)
    return key
  end)
  table.sort(existingDrivers, function(a, b)
    return a < b
  end)

  if table.nkeys(tempDrivers) == 0 then
    ui.text('No specific driver settings yet')
  else
    ui.beginGroup()
    ui.columns(4, false)
    ui.setColumnWidth(0, 25)
    ui.setColumnWidth(1, ui.modernButton and 320 or 270)
    ui.setColumnWidth(2, 70)

    table.forEach(existingDrivers, function(driverName, index)
      local driverEntry = tempDrivers[driverName]
      if driverEntry ~= nil then
        local hidden = driverEntry.hidden
        if hidden then
          ui.pushStyleColor(ui.StyleColor.Text, rgbm.colors.gray)
        end

        ui.alignTextToFramePadding()
        ui.text(index)
        ui.nextColumn()

        ui.alignTextToFramePadding()
        ui.text(driverName)
        ui.nextColumn()

        if ui.checkbox('Hide##' .. driverName, driverEntry.hidden) then
          driverEntry.hidden = not driverEntry.hidden
        end
        ui.nextColumn()

        if hidden then
          ui.popStyleColor()
        end

        -- buttons
        if RemoveButton(driverName) then
          tempDrivers[driverName] = nil
        end

        ui.nextColumn()
      end
    end)

    ui.columns()
    ui.endGroup()
  end

  SpacingY()

  if CopyPaste('drivers', 'Driver definitions', self.driversStore, { 'drivers' }, function()
        self.drivers = self:parseDrivers()
        self:addChangedAttribute('drivers')
      end) then
    return
  end

  -- Add Drivers
  SpacingSeparator()
  ui.setNextItemWidth(ui.availableSpaceX())
  ui.combo('##newdriver', 'Add driver', function()
    local drivers = table.map(table.filter(CARS, function(car)
      return car.isActive and not car.isHidingLabels and not table.contains(existingDrivers, car.driver.name)
    end), function(car)
      return car.driver.name
    end)

    table.sort(drivers, function(a, b)
      return a < b
    end)

    -- draw calculated tags
    table.forEach(drivers, function(driverName)
      if ui.selectable(string.format('%s', driverName)) then
        tempDrivers[driverName] = defaultDriver()
      end
    end)
  end)

  self:setDrivers(tempDrivers)

  -- Reset
  if ui.button('Reset to default') then
    self:setDrivers({})
  end
end

function MapDriverSettings:parseDrivers()
  -- in csp 0.1.79 tryParse returns nil in case of error
  return stringify.tryParse(self.driversStore.drivers, {}, {}) or {}
end

function MapDriverSettings:getDrivers()
  return self.drivers
end

function MapDriverSettings:setDrivers(drivers)
  if not TableCompare(drivers, stringify.tryParse(self.driversStore.drivers, {})) then
    self.drivers = drivers
    self.driversStore.drivers = stringify(drivers)

    self:addChangedAttribute('drivers')
  end
end
